########################################################################
# File name: conversation.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import asyncio
import enum

import aioxmpp.callbacks


class InviteMode(enum.Enum):
    """
    Represent different possible modes for sending an invitation.

    .. attribute:: DIRECT

       The invitation is sent directly to the invitee, without going through a
       service specific to the conversation.

    .. attribute:: MEDIATED

       The invitation is sent indirectly through a service which is providing
       the conversation. Advantages of using this mode include most notably
       that the service can automatically add the invitee to the list of
       allowed participants in configurations where such restrictions exist (or
       deny the request if the inviter does not have the permissions to do so).
    """

    DIRECT = 0
    MEDIATED = 1


class ConversationFeature(enum.Enum):
    """
    Represent individual features of a :term:`Conversation` of a
    :term:`Conversation Implementation`.

    .. seealso::

       The :attr:`.AbstractConversation.features` provides a set of features
       offered by a specific :term:`Conversation`.

    .. attribute:: BAN

       Allows use of :meth:`~.AbstractConversation.ban`.

    .. attribute:: BAN_WITH_KICK

       Explicit support for setting the `request_kick` argument to :data:`True`
       in :meth:`~.AbstractConversation.ban`.

    .. attribute:: INVITE

       Allows use of :meth:`~.AbstractConversation.invite`.

    .. attribute:: INVITE_DIRECT

       Explicit support for the :attr:`~.InviteMode.DIRECT` invite mode when
       calling :meth:`~.AbstractConversation.invite`.

    .. attribute:: INVITE_DIRECT_CONFIGURE

       Explicit support for configuring the conversation to allow the invitee
       to join when using :attr:`~.InviteMode.DIRECT` with
       :meth:`~.AbstractConversation.invite`.

    .. attribute:: INVITE_MEDIATED

       Explicit support for the :attr:`~.InviteMode.MEDIATED` invite mode when
       calling :meth:`~.AbstractConversation.invite`.

    .. attribute:: INVITE_UPGRADE

       Explicit support and requirement for `allow_upgrade` when
       calling :meth:`~.AbstractConversation.invite`.

    .. attribute:: KICK

       Allows use of :meth:`~.AbstractConversation.kick`.

    .. attribute:: LEAVE

       Allows use of :meth:`~.AbstractConversation.leave`.

    .. attribute:: SEND_MESSAGE

       Allows use of :meth:`~.AbstractConversation.send_message`.

    .. attribute:: SEND_MESSAGE_TRACKED

       Allows use of :meth:`~.AbstractConversation.send_message_tracked`.

    .. attribute:: SET_NICK

       Allows use of :meth:`~.AbstractConversation.set_nick`.

    .. attribute:: SET_NICK_OF_OTHERS

       Explicit support for changing the nickname of other members when calling
       :meth:`~.AbstractConversation.set_nick`.

    .. attribute:: SET_TOPIC

       Allows use of :meth:`~.AbstractConversation.set_topic`.

    """

    BAN = 'ban'
    BAN_WITH_KICK = 'ban-with-kick'
    INVITE = 'invite'
    INVITE_DIRECT = 'invite-direct'
    INVITE_DIRECT_CONFIGURE = 'invite-direct-configure'
    INVITE_MEDIATED = 'invite-mediated'
    INVITE_UPGRADE = 'invite-upgrade'
    KICK = 'kick'
    LEAVE = 'leave'
    SEND_MESSAGE = 'send-message'
    SEND_MESSAGE_TRACKED = 'send-message-tracked'
    SET_TOPIC = 'set-topic'
    SET_NICK = 'set-nick'
    SET_NICK_OF_OTHERS = 'set-nick-of-others'


class ConversationState(enum.Enum):
    """
    State of a conversation.

    .. note::

       The members of this enumeration closely mirror the states of :xep:`85`,
       with the addition of the internal :attr:`PENDING` state. The reason is
       that :xep:`85` is a Final Standard and is state-of-the-art for
       conversation states in XMPP.

    .. attribute:: PENDING

       The conversation has been created, possibly automatically, and the
       application has not yet set the conversation state.

    .. attribute:: ACTIVE

       .. epigraph::

          User accepts an initial content message, sends a content message,
          gives focus to the chat session interface (perhaps after being
          inactive), or is otherwise paying attention to the conversation.

          -- from :xep:`85`

    .. attribute:: INACTIVE

       .. epigraph::

          User has not interacted with the chat session interface for an
          intermediate period of time (e.g., 2 minutes).

          -- from :xep:`85`

    .. attribute:: GONE

       .. epigraph::

          User has not interacted with the chat session interface, system, or
          device for a relatively long period of time (e.g., 10 minutes).

          -- from :xep:`85`

    .. attribute:: COMPOSING

       .. epigraph::

          User is actively interacting with a message input interface specific
          to this chat session (e.g., by typing in the input area of a chat
          window).

          -- from :xep:`85`

    .. attribute:: PAUSED

       .. epigraph::

          User was composing but has not interacted with the message input
          interface for a short period of time (e.g., 30 seconds).

          -- from :xep:`85`

    When any of the above states is entered, a notification is sent out to the
    participants of the conversation.
    """

    PENDING = 0
    ACTIVE = 1
    INACTIVE = 2
    GONE = 3
    COMPOSING = 4
    PAUSED = 5


class AbstractConversationMember(metaclass=abc.ABCMeta):
    """
    Represent a member in a :class:`~.AbstractConversation`.

    While all :term:`implementations <Conversation Implementation>` will have
    their own additional attributes, the following attributes must exist on
    all subclasses:

    .. autoattribute:: conversation_jid

    .. autoattribute:: direct_jid

    .. autoattribute:: is_self

    .. autoattribute:: uid
    """

    def __init__(self,
                 conversation_jid,
                 is_self):
        super().__init__()
        self._conversation_jid = conversation_jid
        self._is_self = is_self

    @property
    def direct_jid(self):
        """
        If available, this is the :class:`~aioxmpp.JID` address of the member
        for direct contact, outside of the conversation. It is independent of
        the conversation itself.

        If not available, this attribute reads as :data:`None`.
        """
        return None

    @property
    def conversation_jid(self):
        """
        The :class:`~aioxmpp.JID` of the conversation member relative to the
        conversation.
        """
        return self._conversation_jid

    @property
    def is_self(self):
        """
        True if the member refers to ourselves in the conversation, false
        otherwise.
        """
        return self._is_self

    @abc.abstractproperty
    def uid(self) -> bytes:
        """
        This is a unique ID for the occupant. It can be used across sessions
        and restarts to assert equality between occupants. It is guaranteed
        to be equal if and only if the entity is the same (up to a uncertainty
        caused by the limited length of the unique ID, somewhere in the order
        of ``2**(-120)``).

        The identifier is always a :class:`bytes` and **must** be treated as
        opaque by users. The only guarantee which is given is that its length
        will be less than 4096 bytes.
        """


class AbstractConversation(metaclass=abc.ABCMeta):
    """
    Interface for a conversation.

    .. note::

       All signals may receive additional keyword arguments depending on the
       specific subclass implementing them. Handlers connected to the signals
       **must** support arbitrary keyword arguments.

       To support future extensions to the base specification, subclasses must
       prefix all keyword argument names with a common, short prefix which ends
       with an underscore. For example, a MUC implementation could use
       ``muc_presence``.

       Future extensions to the base class will use either names without
       underscores or the ``base_`` prefix.

    .. note::

       In the same spirit, methods defined on subclasses should use the same
       prefix. However, the base class does not guarantee that it won’t use
       names with underscores in future extensions.

       To prevent collisions, subclasses should avoid the use of prefixes which
       are verbs in the english language.

    Signals:

    .. note::

        The `member` argument common to many signals is never :data:`None` and
        alwyas an instance of a subclass of
        :class:`~.AbstractConversationMember`. However, the `member` may not be
        part of the :attr:`members` of the conversation. For example, it may be
        the :attr:`service_member` object which is never part of
        :attr:`members`. Other cases where a non-member is passed as `member`
        may exist depending on the conversation subclass.

    .. signal:: on_message(msg, member, source, tracker=None, **kwargs)

       A message occured in the conversation.

       :param msg: Message which was received.
       :type msg: :class:`aioxmpp.Message`
       :param member: The member object of the sender.
       :type member: :class:`.AbstractConversationMember`
       :param source: How the message was acquired
       :type source: :class:`~.MessageSource`
       :param tracker: A message tracker which tracks an outbound message.
       :type tracker: :class:`aioxmpp.tracking.MessageTracker`

       This signal is emitted on the following events:

       * A message was sent to the conversation and delivered directly to us.
         This is the classic case of "a message was received". In this case,
         `source` is :attr:`~.MessageSource.STREAM` and `member` is the
         :class:`~.AbstractConversationMember` of the originator.

       * A message was sent from this client. This is the classic case of "a
         message was sent". In this case, `source` is
         :attr:`~.MessageSource.STREAM` and `member` refers to ourselves.

       * A carbon-copy of a message received by another resource of our account
         which belongs to this conversation was received. `source` is
         :attr:`~.MessageSource.CARBONS` and `member` is the
         :class:`~.AbstractConversationMember` of the originator.

       * A carbon-copy of a message sent by another resource of our account was
         sent to this conversation. In this case, `source` is
         :attr:`~.MessageSource.CARBONS` and `member` refers to ourselves.

       Often, you don’t need to distinguish between carbon-copied and
       non-carbon-copied messages.

       All messages which are not handled otherwise (and for example dispatched
       as :meth:`on_state_changed` signals) are dispatched to this event. This
       may include messages not understood and/or which carry no textual
       payload.

       `tracker` is set only for messages sent by the local member. If a
       message is sent from the client without tracking, `tracker` is
       :data:`None`; otherwise, the `tracker` is always set, even for messages
       sent by other clients. It depends on the conversation implementation as
       well as timing in which state a tracker is at the time the event is
       emitted.

    .. signal:: on_state_changed(member, new_state, msg, **kwargs)

       The conversation state of a member has changed.

       :param member: The member object of the member whose state changed.
       :type member: :class:`.AbstractConversationMember`
       :param new_state: The new conversation state of the member.
       :type new_state: :class:`~.ConversationState`
       :param msg: The stanza which conveyed the state change.
       :type msg: :class:`aioxmpp.Message`

       This signal also fires for state changes of the local occupant. The
       exact point at which this signal fires for the local occupant is
       determined by the implementation.

    .. signal:: on_presence_changed(member, resource, presence, **kwargs)

       The presence state of a member has changed.

       :param member: The member object of the affected member.
       :type member: :class:`~.AbstractConversationMember`
       :param resource: The resource of the member which changed presence.
       :type resource: :class:`str` or :data:`None`
       :param presence: The presence stanza
       :type presence: :class:`aioxmpp.Presence`

       If the `presence` stanza affects multiple resources, `resource` holds
       the affected resource and the event is emmited once per affected
       resource.

       However, the `presence` stanza affects only a single resource,
       `resource` is :data:`None`; the affected resource can be extracted from
       the :attr:`~.StanzaBase.from_` of the `presence` stanza in that case.
       This is to help implementations to know whether a bunch of resources was
       shot offline by a single presence (`resource` is not :data:`None`), e.g.
       due to an error or whether a single resource went offline by itself.
       Implementations may want to only show the former case.

       .. note::

          In some implementations, unavailable presence implies that a
          participant leaves the room, in which case :meth:`on_leave` is
          emitted instead.

    .. signal:: on_nick_changed(member, old_nick, new_nick, **kwargs)

       The nickname of a member has changed

       :param member: The member object of the member whose nick has changed.
       :type member: :class:`~.AbstractConversationMember`
       :param old_nick: The old nickname of the member.
       :type old_nick: :class:`str` or :data:`None`
       :param new_nick: The new nickname of the member.
       :type new_nick: :class:`str`

       The new nickname is already set in the `member` object, if the `member`
       object has an accessor for the nickname.

       In some cases, `old_nick` may be :data:`None`. These cases include those
       where it is not trivial for the protocol to actually determine the old
       nickname or where no nickname was set before.

    .. signal:: on_topic_changed(member, new_topic, **kwargs)

        The topic of the conversation has changed.

        :param member: The member object who changed the topic.
        :type member: :class:`~.AbstractConversationMember`
        :param new_topic: The new topic of the conversation.
        :type new_topic: :class:`.LanguageMap`

    .. signal:: on_uid_changed(member, old_uid, **kwargs)

        This rare signal notifies that the
        :attr:`~.AbstractConversationMember.uid` of a member has changed.

        :param member: The member object for which the UID has changed.
        :type member: :class:`~.AbstractConversationMember`
        :param old_uid: The old UID of the member.
        :type old_uid: :class:`bytes`

        The new uid is already available at the members
        :attr:`~.AbstractConversationMember.uid` attribute.

        This signal can only fire for multi-user conversations where the
        visibility of identifying information changes. In many cases, it will
        be irrelevant for the application, but for some use-cases it might be
        important to be able to re-write historical messages to use the new
        uid.

    .. signal:: on_enter()

        The conversation was entered.

        This event is emitted up to once for a :class:`AbstractConversation`.

        One of :meth:`on_enter` and :meth:`on_failure` is emitted exactly
        once for each :class:`AbstractConversation` instance.

        .. seealso::

            :func:`aioxmpp.callbacks.first_signal` can be used nicely to await
            the completion of entering a conversation::

                conv = ... # let this be your conversation
                await first_signal(conv.on_enter, conv.on_failure)
                # await first_signal() will either return None (success) or
                # raise the exception passed to :meth:`on_failure`.

        .. note::

            This and :meth:`on_failure` are the only signals which **must not**
            receive keyword arguments, so that they continue to work with
            :attr:`.AdHocSignal.AUTO_FUTURE` and
            :func:`~.callbacks.first_signal`.

        .. versionadded:: 0.10

    .. signal:: on_failure(exc)

        The conversation could not be entered.

        :param exc: The exception which caused the operation to fail.
        :type exc: :class:`Exception`

        Often, `exc` will be a :class:`aioxmpp.errors.XMPPError` indicating
        an error emitted from an involved server, such as permission problems,
        conflicts or non-existant peers.

        This signal can only be emitted instead of :meth:`on_enter` and not
        after the room has been entered. If the conversation is terminated
        due to a remote cause at a later point, :meth:`on_exit` is used.

        One of :meth:`on_enter` and :meth:`on_failure` is emitted exactly
        once for each :class:`AbstractConversation` instance.

        .. note::

            This and :meth:`on_failure` are the only signals which **must not**
            receive keyword arguments, so that they continue to work with
            :attr:`.AdHocSignal.AUTO_FUTURE` and
            :func:`~.callbacks.first_signal`.

        .. versionadded:: 0.10

    .. signal:: on_join(member, **kwargs)

       A new member has joined the conversation.

       :param member: The member object of the new member.
       :type member: :class:`~.AbstractConversationMember`

       When this signal is called, the `member` is already included in the
       :attr:`members`.

    .. signal:: on_leave(member, **kwargs)

       A member has left the conversation.

       :param member: The member object of the previous member.
       :type member: :class:`~.AbstractConversationMember`

       When this signal is called, the `member` has already been removed from
       the :attr:`members`.

    .. signal:: on_exit(**kwargs)

       The local user has left the conversation.

       When this signal fires, the conversation is defunct in the sense that it
       cannot be used to send messages anymore. A new conversation needs to be
       started.

    Properties:

    .. autoattribute:: features

    .. autoattribute:: jid

    .. autoattribute:: members

    .. autoattribute:: me

    .. autoattribute:: service_member

    Methods:

    .. note::

       See :attr:`features` for discovery of support for individual methods at
       a given conversation instance.

    .. automethod:: ban

    .. automethod:: invite

    .. automethod:: kick

    .. automethod:: leave

    .. automethod:: send_message

    .. automethod:: send_message_tracked

    .. automethod:: set_nick

    .. automethod:: set_topic

    Interface solely for subclasses:

    .. attribute:: _client

       The `client` as passed to the constructor.

    """

    on_message = aioxmpp.callbacks.Signal()
    on_state_changed = aioxmpp.callbacks.Signal()
    on_presence_changed = aioxmpp.callbacks.Signal()
    on_join = aioxmpp.callbacks.Signal()
    on_leave = aioxmpp.callbacks.Signal()
    on_exit = aioxmpp.callbacks.Signal()
    on_failed = aioxmpp.callbacks.Signal()
    on_nick_changed = aioxmpp.callbacks.Signal()
    on_enter = aioxmpp.callbacks.Signal()
    on_failure = aioxmpp.callbacks.Signal()
    on_topic_changed = aioxmpp.callbacks.Signal()

    def __init__(self, service, parent=None, **kwargs):
        super().__init__(**kwargs)
        self._service = service
        self._client = service.client
        self.__parent = parent

    def _not_implemented_error(self, what):
        return NotImplementedError(
            "{} not supported for this type of conversation".format(what)
        )

    @property
    def parent(self):
        """
        The conversation to which this conversation belongs. Read-only.

        When the parent is closed, the sub-conversations are also closed.
        """
        return self.__parent

    @abc.abstractproperty
    def members(self):
        """
        An iterable of members of this conversation.
        """

    @abc.abstractproperty
    def me(self):
        """
        The member representing the local member.
        """

    @abc.abstractproperty
    def jid(self):
        """
        The address of the conversation.
        """

    @property
    def service_member(self):
        """
        The member representing the service on which the conversation is hosted,
        if available.

        This is never included in :attr:`members`. It may be used as member
        argument in events to make it clear that the message originates from
        the service and not an unknown occupant.

        This may be :data:`None`.

        .. versionadded:: 0.10
        """

    @property
    def features(self):
        """
        A set of features supported by this :term:`Conversation`.

        The members of the set are usually drawn from the
        :class:`~.ConversationFeature` :mod:`enumeration <enum>`;
        :term:`Conversation  Implementations <Conversation Implementation>` are
        free to add custom elements from other enumerations to this set.

        Unless stated otherwise, the methods of :class:`~.AbstractConversation`
        and its subclasses always may throw one of the following exceptions,
        **unless** support for those methods is explicitly stated with an
        appropriate :class:`~.ConversationFeature` member in the
        :attr:`features`.

        * :class:`NotImplementedError` if the :term:`Conversation
          Implementation` does not support the method at all.
        * :class:`RuntimeError` if the server does not support the method.
        * :class:`aioxmpp.XMPPCancelError` with ``feature-not-implemented``
          condition.

        *If* support for the method is claimed in :attr:`features`, these
        exceptions **must not** be raised (for the given reason; of course, a
        method may still raise an :class:`aioxmpp.XMPPCancelError` due for
        other conditions such as ``item-not-found``).
        """
        return frozenset()

    def send_message(self, body):
        """
        Send a message to the conversation.

        :param msg: The message to send.
        :type msg: :class:`aioxmpp.Message`
        :return: The stanza token obtained from sending.
        :rtype: :class:`~aioxmpp.stream.StanzaToken`

        The default implementation simply calls :meth:`send_message_tracked`
        and immediately cancels the tracking object, returning only the stanza
        token.

        There is no need to provide proper address attributes on `msg`.
        Implementations will override those attributes with the values
        appropriate for the conversation. Some implementations may allow the
        user to choose a :attr:`~aioxmpp.Message.type_`, but others may simply
        stamp it over.

        Subclasses may override this method with a more specialised
        implementation. Subclasses which do not provide tracked message sending
        **must** override this method to provide untracked message sending.

        .. seealso::

           The corresponding feature is
           :attr:`.ConversationFeature.SEND_MESSAGE`. See :attr:`features` for
           details.

        """
        token, tracker = self.send_message_tracked(body)
        tracker.cancel()
        return token

    @abc.abstractmethod
    def send_message_tracked(self, msg, *, timeout=None):
        """
        Send a message to the conversation with tracking.

        :param msg: The message to send.
        :type msg: :class:`aioxmpp.Message`
        :param timeout: Timeout for the tracking.
        :type timeout: :class:`numbers.RealNumber`, :class:`datetime.timedelta`
                       or :data:`None`
        :raise NotImplementedError: if tracking is not implemented
        :return: The stanza token obtained from sending and the
            :class:`aioxmpp.tracking.MessageTracker` tracking the delivery.
        :rtype: :class:`~aioxmpp.stream.StanzaToken`,
            :class:`~aioxmpp.tracking.MessageTracker`

        There is no need to provide proper address attributes on `msg`.
        Implementations will override those attributes with the values
        appropriate for the conversation. Some implementations may allow the
        user to choose a :attr:`~aioxmpp.Message.type_`, but others may simply
        stamp it over.

        Tracking may not be supported by all implementations, and the degree of
        support varies with implementation. Please check the documentation
        of the respective subclass.

        `timeout` is the number of seconds (or a :class:`datetime.timedelta`
        object which defines the timespan) after which the tracking expires and
        is closed if no response has been received in the mean time. If
        `timeout` is set to :data:`None`, the tracking never expires.

        .. warning::

            Read :ref:`api-tracking-memory`.

        .. seealso::

           The corresponding feature is
           :attr:`.ConversationFeature.SEND_MESSAGE_TRACKED`. See
           :attr:`features` for details.

        """

    @asyncio.coroutine
    def kick(self, member, reason=None):
        """
        Kick a member from the conversation.

        :param member: The member to kick.
        :param reason: A reason to show to the members of the conversation
            including the kicked member.
        :type reason: :class:`str`
        :raises aioxmpp.errors.XMPPError: if the server returned an error for
                                          the kick command.

        .. seealso::

           The corresponding feature is
           :attr:`.ConversationFeature.KICK`. See :attr:`features` for details.
        """
        raise self._not_implemented_error("kicking members")

    @asyncio.coroutine
    def ban(self, member, reason=None, *, request_kick=True):
        """
        Ban a member from re-joining the conversation.

        :param member: The member to ban.
        :param reason: A reason to show to the members of the conversation
            including the banned member.
        :type reason: :class:`str`
        :param request_kick: A flag indicating that the member should be
            removed from the conversation immediately, too.
        :type request_kick: :class:`bool`

        If `request_kick` is true, the implementation attempts to kick the
        member from the conversation, too, if that does not happen
        automatically. There is no guarantee that the member is not removed
        from the conversation even if `request_kick` is false.

        Additional features:

        :attr:`~.ConversationFeature.BAN_WITH_KICK`
           If `request_kick` is true, the member is kicked from the
           conversation.

        .. seealso::

           The corresponding feature for this method is
           :attr:`.ConversationFeature.BAN`. See :attr:`features` for details
           on the semantics of features.
        """
        raise self._not_implemented_error("banning members")

    @asyncio.coroutine
    def invite(self, address, text=None, *,
               mode=InviteMode.DIRECT,
               allow_upgrade=False):
        """
        Invite another entity to the conversation.

        :param address: The address of the entity to invite.
        :type address: :class:`aioxmpp.JID`
        :param text: A reason/accompanying text for the invitation.
        :param mode: The invitation mode to use.
        :type mode: :class:`~.im.InviteMode`
        :param allow_upgrade: Whether to allow creating a new conversation to
            satisfy the invitation.
        :type allow_upgrade: :class:`bool`
        :raises NotImplementedError: if the requested `mode` is not supported
        :raises ValueError: if `allow_upgrade` is false, but a new conversation
            is required.
        :return: The stanza token for the invitation and the possibly new
            conversation object
        :rtype: tuple of :class:`~.StanzaToken` and
            :class:`~.AbstractConversation`

        .. note::

            Even though this is a coroutine, it returns a stanza token. The
            coroutine-ness may be needed to generate the invitation in the
            first place. Sending the actual invitation is done non-blockingly
            and the stanza token for that is returned. To wait until the
            invitation has been sent, unpack the stanza token from the result
            and await it.

        Return the new conversation object to use. In many cases, this will
        simply be the current conversation object, but in some cases (e.g. when
        someone is invited to a one-on-one conversation), a new conversation
        must be created and used.

        If `allow_upgrade` is false and a new conversation would be needed to
        invite an entity, :class:`ValueError` is raised.

        Additional features:

        :attr:`~.ConversationFeature.INVITE_DIRECT`
           Support for :attr:`~.im.InviteMode.DIRECT` mode.

        :attr:`~.ConversationFeature.INVITE_DIRECT_CONFIGURE`
           If a direct invitation is used, the conversation will be configured
           to allow the invitee to join before the invitation is sent. This may
           fail with a :class:`aioxmpp.errors.XMPPError`, in which case the
           error is re-raised and the invitation not sent.

        :attr:`~.ConversationFeature.INVITE_MEDIATED`
           Support for :attr:`~.im.InviteMode.MEDIATED` mode.

        :attr:`~.ConversationFeature.INVITE_UPGRADE`
           If `allow_upgrade` is :data:`True`, an upgrade will be performed and
           a new conversation is returned. If `allow_upgrade` is :data:`False`,
           the invite will fail.

        .. seealso::

           The corresponding feature for this method is
           :attr:`.ConversationFeature.INVITE`. See :attr:`features` for
           details on the semantics of features.
        """
        raise self._not_implemented_error("inviting entities")

    @asyncio.coroutine
    def set_nick(self, new_nickname):
        """
        Change our nickname.

        :param new_nickname: The new nickname for the member.
        :type new_nickname: :class:`str`
        :raises ValueError: if the nickname is not a valid nickname

        Sends the request to change the nickname and waits for the request to
        be sent.

        There is no guarantee that the nickname change will actually be
        applied; listen to the :meth:`on_nick_changed` event.

        Implementations may provide a different method which provides more
        feedback.

        .. seealso::

           The corresponding feature for this method is
           :attr:`.ConversationFeature.SET_NICK`. See :attr:`features` for
           details on the semantics of features.

        """
        raise self._not_implemented_error("changing the nickname")

    @asyncio.coroutine
    def set_topic(self, new_topic):
        """
        Change the (possibly publicly) visible topic of the conversation.

        :param new_topic: The new topic for the conversation.
        :type new_topic: :class:`str`

        Sends the request to change the topic and waits for the request to
        be sent.

        There is no guarantee that the topic change will actually be
        applied; listen to the :meth:`on_topic_chagned` event.

        Implementations may provide a different method which provides more
        feedback.

        .. seealso::

           The corresponding feature for this method is
           :attr:`.ConversationFeature.SET_TOPIC`. See :attr:`features` for
           details on the semantics of features.


        """
        raise self._not_implemented_error("changing the topic")

    @asyncio.coroutine
    def leave(self):
        """
        Leave the conversation.

        .. seealso::

           The corresponding feature is
           :attr:`.ConversationFeature.LEAVE`. See :attr:`features` for
           details.
        """


class AbstractConversationService(metaclass=abc.ABCMeta):
    """
    Abstract base class for
    :term:`Conversation Services <Conversation Service>`.

    Useful implementations:

    .. autosummary::

        aioxmpp.im.p2p.Service
        aioxmpp.muc.MUCClient

    In general, conversation services should provide a method (*not* a coroutine
    method) to start a conversation using the service. That method should return
    the fresh :class:`~.AbstractConversation` object immediately and start
    possibly needed background tasks to actually initiate the conversation. The
    caller should use the :meth:`~.AbstractConversation.on_enter` and
    :meth:`~.AbstractConversation.on_failure` signals to be notified of the
    result of the join operation.

    Signals:

    .. signal:: on_conversation_new(conversation)

        Fires when a new conversation is created in the service.

        :param conversation: The new conversation.
        :type conversation: :class:`AbstractConversation`

        .. seealso::

            :meth:`.ConversationService.on_conversation_added`
                is a signal shared among all :term:`Conversation
                Implementations <Conversation Implementation>` which gets
                emitted whenever a new conversation is added. If you need all
                conversations, that is the signal to listen for.

    .. signal:: on_spontaneous_conversation(conversation)

        Like :meth:`on_conversation_new`, but is only emitted for conversations
        which are created without local interaction.

        :param conversation: The new conversation.
        :type conversation: :class:`AbstractConversation`

        .. versionadded:: 0.10

    """

    on_conversation_new = aioxmpp.callbacks.Signal()
    on_spontaneous_conversation = aioxmpp.callbacks.Signal()



########################################################################
# File name: query_versions.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import argparse
import asyncio
import itertools
import sys

import aioxmpp
import aioxmpp.disco
import aioxmpp.errors
import aioxmpp.version
import aioxmpp.xso

from framework import Example, exec_example


if not hasattr(asyncio, "ensure_future"):
    asyncio.ensure_future = getattr(asyncio, "async")


class SoftwareVersions(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        self.argparse.add_argument(
            "--timeout",
            type=float,
            help="Maximum time (in seconds) to wait for a response "
            "(default: 20s)",
            default=20,
        )

        group = self.argparse.add_mutually_exclusive_group(required=True)

        group.add_argument(
            "-f", "--from-file",
            type=argparse.FileType("r"),
            dest="from_file",
            help="Read the JIDs from a file, one line per JID.",
            default=None,
        )

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        group.add_argument(
            "-t", "--targets",
            nargs="+",
            dest="jids",
            help="The JIDs to query as command-line arguments",
            default=[],
            type=jid,
        )

    def configure(self):
        super().configure()

        self.jids = list(self.args.jids)
        if self.args.from_file:
            with self.args.from_file as f:
                for line in f:
                    self.jids.append(aioxmpp.JID.fromstr(line[:-1]))

        self.timeout = self.args.timeout

    @asyncio.coroutine
    def run_example(self):
        self.stop_event = self.make_sigint_event()
        yield from super().run_example()

    def format_version(self, version_xso):
        return "name={!r} version={!r} os={!r}".format(
            version_xso.name,
            version_xso.version,
            version_xso.os,
        )

    @asyncio.coroutine
    def run_simple_example(self):
        tasks = []
        stream = self.client.stream
        for jid in self.jids:
            tasks.append(
                asyncio.ensure_future(aioxmpp.version.query_version(
                    stream,
                    jid,
                ))
            )

        gather_task = asyncio.ensure_future(
            asyncio.wait(
                tasks,
                return_when=asyncio.ALL_COMPLETED,
            )
        )

        cancel_fut = asyncio.ensure_future(self.stop_event.wait())

        yield from asyncio.wait(
            [
                gather_task,
                cancel_fut,
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )

        for target, fut in zip(self.jids, tasks):
            if not fut.done():
                fut.cancel()
                continue

            if fut.exception():
                print("{} failed: {}".format(target, fut.exception()),
                      file=sys.stderr)
                continue

            print("{}: {}".format(target, self.format_version(fut.result())))

        if not cancel_fut.done():
            cancel_fut.cancel()


if __name__ == "__main__":
    exec_example(SoftwareVersions())



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import typing

import aioxmpp
import aioxmpp.disco
import aioxmpp.errors
import aioxmpp.service

from aioxmpp.utils import namespaces

from . import xso as version_xso


class VersionServer(aioxmpp.service.Service):
    """
    :class:`~aioxmpp.service.Service` which handles inbound :xep:`92` Software
    Version requests.

    .. warning::

        Do **not** depend on this service in another service. This service
        exposes possibily private or sensitive information over the XMPP
        network without any filtering. Implicitly summoning this service via
        a dependency is thus discouraged.

        *If* you absolutely need to do this for the implementation of another
        published XEP, please file an issue against :mod:`aioxmpp` so that we
        can work out a good solution.

    .. warning::

        This service does answer version queries, no matter who asks. This may
        not be desirable, in which case this service is not for you.

    .. seealso::

        :func:`~.version.query_version`
            for a function to obtain another entities software version.

    .. note::

        By default, this service does not reply to version queries. The
        :attr:`name` attribute needs to be set first.

    The response can be configured with the following attributes:

    .. autoattribute:: name
        :annotation: = None

    .. autoattribute:: version
        :annotation: = None

    .. autoattribute:: os
        :annotation: = distro.name() or platform.system()
    """

    ORDER_AFTER = [
        aioxmpp.disco.DiscoServer,
    ]

    disco_feature = aioxmpp.disco.register_feature(
        "jabber:iq:version",
    )

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)
        try:
            import distro
        except ImportError:
            import platform
            self._os = platform.system()
        else:
            self._os = distro.name()

        self._name = None
        self._version = None

    @property
    def os(self) -> typing.Optional[str]:
        """
        The operating system of this entity.

        Defaults to :func:`distro.name` or :func:`platform.system` (if
        :mod:`distro` is not available).

        This attribute can be set to :data:`None` or deleted to prevent
        inclusion of the OS element in the reply.
        """
        return self._os

    @os.setter
    def os(self, value: typing.Optional[str]):
        if value is None:
            self._os = None
        else:
            self._os = str(value)

    @os.deleter
    def os(self):
        self._os = None

    @property
    def name(self) -> typing.Optional[str]:
        """
        The software name of this entity.

        Defaults to :data:`None`.

        If this attribute is :data:`None`, version requests are not answered
        but fail with a ``service-unavailable`` error.
        """
        return self._name

    @name.setter
    def name(self, value: typing.Optional[str]):
        if value is None:
            self._name = None
        else:
            self._name = str(value)

    @name.deleter
    def name(self):
        self._name = None

    @property
    def version(self) -> typing.Optional[str]:
        """
        The software version of this entity.

        Defaults to :data:`None`.

        If this attribute is :data:`None` or the empty string, the version will
        be shown as ``"unspecified"`` to other entities. This can be used to
        avoid disclosing the specific version of the software.
        """
        return self._version

    @version.setter
    def version(self, value: typing.Optional[str]):
        if value is None:
            self._version = None
        else:
            self._version = str(value)

    @version.deleter
    def version(self):
        self._version = None

    @aioxmpp.service.iq_handler(aioxmpp.IQType.GET,
                                version_xso.Query)
    @asyncio.coroutine
    def handle_query(self, iq: aioxmpp.IQ) -> version_xso.Query:
        if self._name is None:
            raise aioxmpp.errors.XMPPCancelError(
                aioxmpp.errors.ErrorCondition.SERVICE_UNAVAILABLE,
            )

        result = version_xso.Query()
        result.name = self._name
        result.os = self._os
        result.version = self._version or "unspecified"
        return result


@asyncio.coroutine
def query_version(stream: aioxmpp.stream.StanzaStream,
                  target: aioxmpp.JID) -> version_xso.Query:
    """
    Query the software version of an entity.

    :param stream: A stanza stream to send the query on.
    :type stream: :class:`aioxmpp.stream.StanzaStream`
    :param target: The address of the entity to query.
    :type target: :class:`aioxmpp.JID`
    :raises OSError: if a connection issue occured before a reply was received
    :raises aioxmpp.errors.XMPPError: if an XMPP error was returned instead
        of a reply.
    :rtype: :class:`aioxmpp.version.xso.Query`
    :return: The response from the peer.

    The response is returned as :class:`~aioxmpp.version.xso.Query` object. The
    attributes hold the data returned by the peer. Each attribute may be
    :data:`None` if the peer chose to omit that information. In an extreme case,
    all attributes are :data:`None`.
    """

    return (yield from stream.send(
        aioxmpp.IQ(
            type_=aioxmpp.IQType.GET,
            to=target,
            payload=version_xso.Query(),
        )
    ))



########################################################################
# File name: test_stanza.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import contextlib
import enum
import io
import itertools
import unittest
import unittest.mock

import aioxmpp.xso as xso
import aioxmpp.stanza as stanza
import aioxmpp.structs as structs
import aioxmpp.errors as errors
import aioxmpp.xml

from aioxmpp.utils import namespaces


TEST_FROM = structs.JID.fromstr("foo@example.test")
TEST_TO = structs.JID.fromstr("bar@example.test")


@stanza.IQ.as_payload_class
class TestPayload(xso.XSO):
    TAG = "foo", "bar"

    def __repr__(self):
        return "foobar"


class TestStanzaBase(unittest.TestCase):
    class FakeStanza(stanza.StanzaBase, protect=False):
        pass

    def test_declare_ns(self):
        self.assertDictEqual(
            stanza.StanzaBase.DECLARE_NS,
            {}
        )

    def test_from_attr(self):
        self.assertIsInstance(
            stanza.StanzaBase.from_,
            xso.Attr)
        self.assertEqual(
            (None, "from"),
            stanza.StanzaBase.from_.tag)
        self.assertIsInstance(
            stanza.StanzaBase.from_.type_,
            xso.JID)

    def test_to_attr(self):
        self.assertIsInstance(
            stanza.StanzaBase.to,
            xso.Attr)
        self.assertEqual(
            (None, "to"),
            stanza.StanzaBase.to.tag)
        self.assertIsInstance(
            stanza.StanzaBase.to.type_,
            xso.JID)

    def test_lang_attr(self):
        self.assertIsInstance(
            stanza.StanzaBase.lang,
            xso.LangAttr)

    def test_error_attr(self):
        self.assertIsInstance(
            stanza.StanzaBase.error,
            xso.Child)
        self.assertIs(stanza.StanzaBase.error.default, None)

    def test_autoset_id_generates_random_str_on_unset(self):
        s = self.FakeStanza()
        s.autoset_id()
        id1 = s.id_
        self.assertTrue(id1.startswith("x"))
        self.assertTrue(s.id_)
        del s.id_
        s.autoset_id()
        self.assertTrue(s.id_)
        self.assertNotEqual(id1, s.id_)
        self.assertIsInstance(s.id_, str)
        self.assertTrue(s.id_.startswith("x"))

        # ensure that there are not too many A chars (i.e. zero bits)
        self.assertLess(sum(1 for c in id1 if c == "A"), 5)

    def test_autoset_id_generates_random_str_on_None(self):
        s = self.FakeStanza()
        s.id_ = None
        s.autoset_id()
        id1 = s.id_
        self.assertTrue(id1.startswith("x"))
        self.assertTrue(s.id_)
        del s.id_
        s.autoset_id()
        self.assertTrue(s.id_)
        self.assertNotEqual(id1, s.id_)
        self.assertIsInstance(s.id_, str)
        self.assertTrue(s.id_.startswith("x"))

        # ensure that there are not too many A chars (i.e. zero bits)
        self.assertLess(sum(1 for c in id1 if c == "A"), 5)

    def test_autoset_id_does_not_override(self):
        s = self.FakeStanza()
        s.id_ = "foo"
        s.autoset_id()
        self.assertEqual("foo", s.id_)

    def test_init(self):
        id_ = "someid"

        s = self.FakeStanza(
            from_=TEST_FROM,
            to=TEST_TO,
            id_=id_)
        self.assertEqual(
            TEST_FROM,
            s.from_)
        self.assertEqual(
            TEST_TO,
            s.to)
        self.assertEqual(
            id_,
            s.id_)

    def test_xso_error_handler_raises_StanzaError(self):
        s = stanza.StanzaBase()
        with self.assertRaisesRegex(
                stanza.StanzaError,
                "failed to parse stanza") as ctx:
            s.xso_error_handler(
                unittest.mock.sentinel.descriptor,
                unittest.mock.sentinel.ev_args,
                unittest.mock.sentinel.exc_info,
            )

        self.assertIs(
            ctx.exception.ev_args,
            unittest.mock.sentinel.ev_args,
        )

        self.assertIs(
            ctx.exception.descriptor,
            unittest.mock.sentinel.descriptor,
        )

        self.assertIs(
            ctx.exception.partial_obj,
            s
        )


class TestBody(unittest.TestCase):
    def test_tag(self):
        self.assertEqual(
            (namespaces.client, "body"),
            stanza.Body.TAG)

    def test_lang_attr(self):
        self.assertIsInstance(
            stanza.Body.lang,
            xso.LangAttr)

    def test_text_attr(self):
        self.assertIsInstance(
            stanza.Body.text,
            xso.Text)


class TestSubject(unittest.TestCase):
    def test_tag(self):
        self.assertEqual(
            (namespaces.client, "subject"),
            stanza.Subject.TAG)

    def test_lang_attr(self):
        self.assertIsInstance(
            stanza.Subject.lang,
            xso.LangAttr)

    def test_text_attr(self):
        self.assertIsInstance(
            stanza.Subject.text,
            xso.Text)


class TestMessage(unittest.TestCase):
    def test_inheritance(self):
        self.assertTrue(issubclass(
            stanza.Message,
            stanza.StanzaBase))

    def test_unknown_child_policy(self):
        self.assertEqual(
            stanza.Message.UNKNOWN_CHILD_POLICY,
            xso.UnknownChildPolicy.DROP
        )

    def test_id_attr(self):
        self.assertIsInstance(
            stanza.Message.id_,
            xso.Attr)
        self.assertEqual(
            (None, "id"),
            stanza.Message.id_.tag)
        self.assertIs(stanza.Message.id_.default, None)

    def test_tag(self):
        self.assertEqual(
            ("jabber:client", "message"),
            stanza.Message.TAG)

    def test_type_attr(self):
        self.assertIsInstance(
            stanza.Message.type_,
            xso.Attr
        )
        self.assertEqual(
            (None, "type"),
            stanza.Message.type_.tag
        )
        self.assertIsInstance(
            stanza.Message.type_.type_,
            xso.EnumCDataType,
        )
        self.assertIs(
            stanza.Message.type_.type_.enum_class,
            structs.MessageType,
        )
        self.assertEqual(
            stanza.Message.type_.default,
            structs.MessageType.NORMAL,
        )

    def test_ext_attr(self):
        self.assertIsInstance(
            stanza.Message.ext,
            xso.ChildMap)

    def test_body_attr(self):
        self.assertIsInstance(
            stanza.Message.body,
            xso.ChildTextMap)
        self.assertSetEqual(
            {stanza.Body},
            set(stanza.Message.body._classes)
        )

    def test_subject_attr(self):
        self.assertIsInstance(
            stanza.Message.subject,
            xso.ChildTextMap)
        self.assertSetEqual(
            {stanza.Subject},
            set(stanza.Message.subject._classes)
        )

    def test_thread_attr(self):
        self.assertIsInstance(
            stanza.Message.thread,
            xso.Child)
        self.assertSetEqual(
            {stanza.Thread},
            set(stanza.Message.thread._classes)
        )

    def test_init(self):
        s = stanza.Message(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.MessageType.GROUPCHAT,
        )
        self.assertEqual(
            TEST_FROM,
            s.from_
        )
        self.assertEqual(
            structs.MessageType.GROUPCHAT,
            s.type_
        )

    def test_reject_init_without_type(self):
        with self.assertRaisesRegex(TypeError, "type_"):
            stanza.Message()

    def test_make_reply(self):
        s = stanza.Message(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.MessageType.GROUPCHAT,
        )
        r = s.make_reply()
        self.assertEqual(
            r.type_,
            s.type_)
        self.assertEqual(
            TEST_FROM,
            r.to)
        self.assertEqual(
            TEST_TO,
            r.from_)
        self.assertIsNone(r.id_)

    def test_make_error(self):
        e = stanza.Error(
            condition=errors.ErrorCondition.FEATURE_NOT_IMPLEMENTED
        )
        s = stanza.Message(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.MessageType.GROUPCHAT
        )
        r = s.make_error(e)

        self.assertIsInstance(r, stanza.Message)

        self.assertEqual(
            r.type_,
            structs.MessageType.ERROR)
        self.assertEqual(
            TEST_FROM,
            r.to)
        self.assertEqual(
            TEST_TO,
            r.from_)
        self.assertEqual(
            s.id_,
            r.id_)

    def test_repr(self):
        s = stanza.Message(from_=TEST_FROM,
                           to=TEST_TO,
                           id_="someid",
                           type_=structs.MessageType.GROUPCHAT)
        self.assertEqual(
            "<message from='foo@example.test' to='bar@example.test'"
            " id='someid' type=<MessageType.GROUPCHAT: 'groupchat'>>",
            repr(s)
        )

    def test_repr_works_with_mostly_uninitialised_attributes(self):
        s = stanza.Message.__new__(stanza.Message)
        self.assertEqual(
            repr(s),
            "<message from=None to=None id=None type=<MessageType.NORMAL: 'normal'>>"
        )

    def test_repr_works_with_incomplete_attributes(self):
        s = stanza.Message.__new__(stanza.Message)
        stanza.Message.from_.mark_incomplete(s)
        stanza.Message.to.mark_incomplete(s)
        stanza.Message.type_.mark_incomplete(s)
        stanza.Message.id_.mark_incomplete(s)
        self.assertEqual(
            repr(s),
            "<message from=<incomplete> to=<incomplete> "
            "id=<incomplete> type=<incomplete>>"
        )

    def test_random_type_is_equal_to_normal(self):
        buf = io.BytesIO(b"<message xmlns='jabber:client' type='fnord'/>")
        s = aioxmpp.xml.read_single_xso(buf, stanza.Message)
        self.assertIs(s.type_, structs.MessageType.NORMAL)

    def test_absent_type_is_normal(self):
        buf = io.BytesIO(b"<message xmlns='jabber:client'/>")
        s = aioxmpp.xml.read_single_xso(buf, stanza.Message)
        self.assertIs(s.type_, structs.MessageType.NORMAL)


class TestStatus(unittest.TestCase):
    def test_tag(self):
        self.assertEqual(
            (namespaces.client, "status"),
            stanza.Status.TAG)

    def test_lang_attr(self):
        self.assertIsInstance(
            stanza.Status.lang,
            xso.LangAttr)

    def test_text_attr(self):
        self.assertIsInstance(
            stanza.Status.text,
            xso.Text)


class TestPresence(unittest.TestCase):
    def test_inheritance(self):
        self.assertIsInstance(
            stanza.Presence(),
            stanza.StanzaBase)

    def test_id_attr(self):
        self.assertIsInstance(
            stanza.Presence.id_,
            xso.Attr)
        self.assertEqual(
            (None, "id"),
            stanza.Presence.id_.tag)
        self.assertIs(stanza.Presence.id_.default, None)

    def test_tag(self):
        self.assertEqual(
            ("jabber:client", "presence"),
            stanza.Presence.TAG)

    def test_type_attr(self):
        self.assertIsInstance(
            stanza.Presence.type_,
            xso.Attr,
        )
        self.assertEqual(
            (None, "type"),
            stanza.Presence.type_.tag,
        )
        self.assertIsInstance(
            stanza.Presence.type_.type_,
            xso.EnumCDataType,
        )
        self.assertIs(
            stanza.Presence.type_.type_.enum_class,
            structs.PresenceType,
        )
        self.assertIs(
            stanza.Presence.type_.default,
            structs.PresenceType.AVAILABLE,
        )

    def test_show_attr(self):
        self.assertIsInstance(
            stanza.Presence.show,
            xso.ChildText,
        )
        self.assertEqual(
            (namespaces.client, "show"),
            stanza.Presence.show.tag,
        )
        self.assertIsInstance(
            stanza.Presence.show.type_,
            xso.EnumCDataType,
        )
        self.assertIs(
            stanza.Presence.show.type_.enum_class,
            structs.PresenceShow,
        )
        self.assertIs(
            stanza.Presence.show.default,
            structs.PresenceShow.NONE,
        )

    def test_status_attr(self):
        self.assertIsInstance(
            stanza.Presence.status,
            xso.ChildTextMap)
        self.assertSetEqual(
            {stanza.Status},
            set(stanza.Presence.status._classes)
        )

    def test_priority_attr(self):
        self.assertIsInstance(
            stanza.Presence.priority,
            xso.ChildText)
        self.assertEqual(
            (namespaces.client, "priority"),
            stanza.Presence.priority.tag
        )
        self.assertIsInstance(
            stanza.Presence.priority.type_,
            xso.Integer
        )
        self.assertEqual(
            0,
            stanza.Presence.priority.default
        )

    def test_ext_attr(self):
        self.assertIsInstance(
            stanza.Presence.ext,
            xso.ChildMap)

    def test_error_attr(self):
        self.assertIsInstance(
            stanza.Presence.error,
            xso.Child)

    def test_init(self):
        s = stanza.Presence(
            from_=TEST_FROM,
            type_=structs.PresenceType.PROBE,
            show=structs.PresenceShow.AWAY,
        )
        self.assertEqual(
            TEST_FROM,
            s.from_
        )
        self.assertEqual(
            structs.PresenceType.PROBE,
            s.type_
        )
        self.assertEqual(
            structs.PresenceShow.AWAY,
            s.show,
        )

    def test_init_compat(self):
        s = stanza.Presence(
            from_=TEST_FROM,
            type_=structs.PresenceType.PROBE,
            show="xa",
        )
        self.assertEqual(
            TEST_FROM,
            s.from_
        )
        self.assertEqual(
            structs.PresenceType.PROBE,
            s.type_
        )
        self.assertEqual(
            structs.PresenceShow.XA,
            s.show,
        )

    def test_default(self):
        s = stanza.Presence()
        self.assertEqual(
            s.type_,
            structs.PresenceType.AVAILABLE,
        )
        self.assertEqual(s.show, structs.PresenceShow.NONE)

    def test_make_error(self):
        e = stanza.Error(
            condition=errors.ErrorCondition.GONE
        )
        s = stanza.Presence(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.PresenceType.UNAVAILABLE,
        )
        r = s.make_error(e)

        self.assertIsInstance(r, stanza.Presence)

        self.assertEqual(
            r.type_,
            structs.PresenceType.ERROR
        )
        self.assertEqual(
            TEST_FROM,
            r.to
        )
        self.assertEqual(
            TEST_TO,
            r.from_
        )
        self.assertEqual(
            s.id_,
            r.id_
        )

    def test_repr(self):
        s = stanza.Presence(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.PresenceType.PROBE,
        )
        self.assertEqual(
            "<presence from='foo@example.test' to='bar@example.test'"
            " id='someid' type=<PresenceType.PROBE: 'probe'>>",
            repr(s)
        )
        s = stanza.Presence(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.PresenceType.AVAILABLE
        )
        self.assertEqual(
            "<presence from='foo@example.test' to='bar@example.test'"
            " id='someid' type=<PresenceType.AVAILABLE: None>>",
            repr(s)
        )

    def test_collector(self):
        self.assertIsInstance(
            stanza.Presence.unhandled_children,
            xso.Collector
        )

    def test_repr_works_with_mostly_uninitialised_attributes(self):
        s = stanza.Presence.__new__(stanza.Presence)
        self.assertEqual(
            repr(s),
            "<presence from=None to=None id=None type=<PresenceType.AVAILABLE: None>>"
        )

    def test_repr_works_with_incomplete_attributes(self):
        s = stanza.Presence.__new__(stanza.Presence)
        stanza.Presence.from_.mark_incomplete(s)
        stanza.Presence.to.mark_incomplete(s)
        stanza.Presence.id_.mark_incomplete(s)
        stanza.Presence.type_.mark_incomplete(s)
        self.assertEqual(
            repr(s),
            "<presence from=<incomplete> to=<incomplete> "
            "id=<incomplete> type=<incomplete>>"
        )

    def test_empty_show_is_equivalent_to_no_show(self):
        buf = io.BytesIO(b"<presence xmlns='jabber:client'><show/></presence>")
        s = aioxmpp.xml.read_single_xso(buf, stanza.Presence)
        self.assertIs(s.show, structs.PresenceShow.NONE)

    def test_absent_show(self):
        buf = io.BytesIO(b"<presence xmlns='jabber:client'/>")
        s = aioxmpp.xml.read_single_xso(buf, stanza.Presence)
        self.assertIs(s.show, structs.PresenceShow.NONE)


class TestError(unittest.TestCase):
    def test_declare_ns(self):
        self.assertDictEqual(
            stanza.Error.DECLARE_NS,
            {}
        )

    def test_tag(self):
        self.assertEqual(
            ("jabber:client", "error"),
            stanza.Error.TAG)

    def test_unknown_child_policy(self):
        self.assertIs(
            stanza.Error.UNKNOWN_CHILD_POLICY,
            xso.UnknownChildPolicy.DROP
        )

    def test_unknown_attr_policy(self):
        self.assertIs(
            stanza.Error.UNKNOWN_ATTR_POLICY,
            xso.UnknownAttrPolicy.DROP
        )

    def test_type_attr(self):
        self.assertIsInstance(
            stanza.Error.type_,
            xso.Attr,
        )
        self.assertEqual(
            (None, "type"),
            stanza.Error.type_.tag,
        )
        self.assertIsInstance(
            stanza.Error.type_.type_,
            xso.EnumCDataType,
        )
        self.assertIs(
            stanza.Error.type_.type_.enum_class,
            structs.ErrorType,
        )

    def test_condition_obj_attr(self):
        self.assertIsInstance(
            stanza.Error.condition_obj,
            xso.Child,
        )
        self.assertCountEqual(
            [
                member.xso_class
                for member in errors.ErrorCondition
            ],
            stanza.Error.condition_obj._classes,
        )
        self.assertTrue(stanza.Error.condition_obj.required)

    def test_initialises_with_undefined_condition(self):
        e = stanza.Error()
        self.assertIsInstance(
            e.condition_obj,
            errors.ErrorCondition.UNDEFINED_CONDITION.xso_class,
        )

    @unittest.skipIf(aioxmpp.version_info >= (1, 0, 0),
                     "does not apply to this version of aioxmpp")
    def test_init_works_with_tuple(self):
        with self.assertWarnsRegex(
                DeprecationWarning,
                r"as of aioxmpp 1\.0, error conditions must be members of the "
                r"aioxmpp\.ErrorCondition enumeration") as ctx:
            e = stanza.Error(
                errors.ErrorCondition.REMOTE_SERVER_NOT_FOUND.value
            )

        self.assertEqual(
            e.condition,
            errors.ErrorCondition.REMOTE_SERVER_NOT_FOUND,
        )

        self.assertTrue(ctx.filename.endswith("test_stanza.py"))

    def test_init_works_with_xso(self):
        condition_obj = errors.ErrorCondition.GONE.to_xso()
        condition_obj.new_address = "foo"

        e = stanza.Error(
            condition_obj
        )

        self.assertIs(e.condition_obj, condition_obj)

    def test_condition_reflects_enum_member_of_object_after_init(self):
        e = stanza.Error()

        self.assertEqual(
            errors.ErrorCondition.UNDEFINED_CONDITION,
            e.condition,
        )

    def test_condition_reflects_enum_member_of_object_after_change(self):
        e = stanza.Error()
        e.condition_obj = errors.ErrorCondition.BAD_REQUEST.xso_class()

        self.assertEqual(
            errors.ErrorCondition.BAD_REQUEST,
            e.condition,
        )

    def test_setting_condition_replaces_object(self):
        e = stanza.Error()
        e.condition = errors.ErrorCondition.UNDEFINED_CONDITION

        self.assertEqual(
            e.condition,
            errors.ErrorCondition.UNDEFINED_CONDITION
        )

        self.assertIsInstance(
            e.condition_obj,
            errors.ErrorCondition.UNDEFINED_CONDITION.xso_class,
        )

    def test_setting_condition_keeps_object_if_condition_matches(self):
        e = stanza.Error()
        old = e.condition_obj
        e.condition = errors.ErrorCondition.UNDEFINED_CONDITION
        self.assertIs(e.condition_obj, old)

    @unittest.skipIf(aioxmpp.version_info >= (1, 0, 0),
                     "does not apply to this version of aioxmpp")
    def test_accepts_tuple_instead_of_enum_for_condition_and_warns(self):
        e = stanza.Error()
        with self.assertWarnsRegex(
                DeprecationWarning,
                r"as of aioxmpp 1\.0, error conditions must be members of the "
                r"aioxmpp\.ErrorCondition enumeration") as ctx:
            e.condition = errors.ErrorCondition.BAD_REQUEST.value

        self.assertEqual(
            errors.ErrorCondition.BAD_REQUEST,
            e.condition,
        )

        self.assertIsInstance(
            e.condition_obj,
            errors.ErrorCondition.BAD_REQUEST.xso_class,
        )

        self.assertTrue(ctx.filename.endswith("test_stanza.py"))

    def test_rejects_xso_for_condition(self):
        e = stanza.Error()

        with self.assertRaises(ValueError):
            e.condition = errors.ErrorCondition.BAD_REQUEST.to_xso()

    def test_application_condition_attr(self):
        self.assertIsInstance(
            stanza.Error.application_condition,
            xso.Child)
        self.assertFalse(stanza.Error.application_condition.required)

    def test_from_exception(self):
        exc = errors.XMPPWaitError(
            condition=errors.ErrorCondition.ITEM_NOT_FOUND,
            text="foobar"
        )
        obj = stanza.Error.from_exception(exc)
        self.assertEqual(
            structs.ErrorType.WAIT,
            obj.type_
        )
        self.assertEqual(
            errors.ErrorCondition.ITEM_NOT_FOUND,
            obj.condition
        )
        self.assertEqual(
            "foobar",
            obj.text
        )

    def test_to_exception(self):
        types = {
            structs.ErrorType.MODIFY: errors.XMPPModifyError,
            structs.ErrorType.CANCEL: errors.XMPPCancelError,
            structs.ErrorType.AUTH: errors.XMPPAuthError,
            structs.ErrorType.WAIT: errors.XMPPWaitError,
            structs.ErrorType.CONTINUE: errors.XMPPContinueError,
        }
        conditions = [
            errors.ErrorCondition.BAD_REQUEST,
            errors.ErrorCondition.UNDEFINED_CONDITION,
        ]
        texts = [
            "foo",
            "bar",
            None,
        ]

        for (type_name, cls), condition, text in itertools.product(
                types.items(),
                conditions,
                texts):
            obj = stanza.Error(
                type_=type_name,
                condition=condition,
                text=text
            )
            exc = obj.to_exception()
            self.assertIsInstance(
                exc,
                cls
            )
            self.assertEqual(
                condition,
                exc.condition
            )
            self.assertIs(
                exc.condition_obj,
                obj.condition_obj,
            )
            self.assertEqual(
                text,
                exc.text
            )

    def test_to_exception_with_application_condition(self):
        cond = unittest.mock.Mock(["to_exception"])

        obj = stanza.Error(
            type_=structs.ErrorType.CONTINUE,
            condition=errors.ErrorCondition.UNDEFINED_CONDITION
        )
        obj.application_condition = cond
        cond.to_exception.return_value = Exception()

        result = obj.to_exception()

        self.assertSequenceEqual(
            cond.mock_calls,
            [
                unittest.mock.call.to_exception(obj.type_)
            ]
        )

        self.assertEqual(result, cond.to_exception())

    def test_to_exception_with_application_condition_only_if_cond_supports(self):
        cond = unittest.mock.Mock(["TAG"])
        cond.TAG = ("foo", "bar")

        obj = stanza.Error(
            type_=structs.ErrorType.CONTINUE,
            condition=errors.ErrorCondition.UNDEFINED_CONDITION
        )
        obj.application_condition = cond

        result = obj.to_exception()

        self.assertIsInstance(
            result,
            errors.XMPPContinueError
        )

        self.assertEqual(
            result.application_defined_condition,
            obj.application_condition,
        )

        self.assertSequenceEqual(
            cond.mock_calls,
            [
            ]
        )

    def test_override_with_default_exception_if_result_of_app_cond_is_no_exception(self):
        cond = unittest.mock.Mock(["to_exception", "TAG"])
        cond.TAG = ("foo", "bar")

        obj = stanza.Error(
            type_=structs.ErrorType.CONTINUE,
            condition=errors.ErrorCondition.UNDEFINED_CONDITION
        )
        obj.application_condition = cond
        cond.to_exception.return_value = object()

        result = obj.to_exception()

        self.assertIsInstance(
            result,
            errors.XMPPContinueError
        )

        self.assertSequenceEqual(
            cond.mock_calls,
            [
                unittest.mock.call.to_exception(obj.type_)
            ]
        )

    def test_from_exception_with_application_condition(self):
        @stanza.Error.as_application_condition
        class Foo(xso.XSO):
            TAG = ("uri:foo", "test_from_exception_with_application_condition")

        obj = Foo()

        exc = errors.XMPPAuthError(
            errors.ErrorCondition.NOT_AUTHORIZED,
            application_defined_condition=obj
        )

        err = stanza.Error.from_exception(exc)

        self.assertIs(
            err.application_condition,
            obj,
        )

    def test_repr(self):
        obj = stanza.Error()
        self.assertEqual(
            "<undefined-condition type=<ErrorType.CANCEL: 'cancel'>>",
            repr(obj)
        )
        obj = stanza.Error(
            type_=structs.ErrorType.MODIFY,
            condition=errors.ErrorCondition.BAD_REQUEST,
            text="foobar"
        )
        self.assertEqual(
            "<bad-request type=<ErrorType.MODIFY: 'modify'> text='foobar'>",
            repr(obj)
        )

    def test_as_application_condition(self):
        @stanza.Error.as_application_condition
        class Foo(xso.XSO):
            TAG = ("uri:foo", "test_as_payload_class")

        self.assertIn(Foo.TAG, stanza.Error.CHILD_MAP)
        self.assertIs(
            stanza.Error.CHILD_MAP[Foo.TAG],
            stanza.Error.application_condition.xq_descriptor
        )


class TestIQ(unittest.TestCase):
    def test_inheritance(self):
        self.assertTrue(issubclass(
            stanza.IQ,
            stanza.StanzaBase))

    def test_unknown_child_policy(self):
        self.assertEqual(
            stanza.IQ.UNKNOWN_CHILD_POLICY,
            xso.UnknownChildPolicy.FAIL
        )

    def test_id_attr(self):
        self.assertIsInstance(
            stanza.IQ.id_,
            xso.Attr)
        self.assertEqual(
            (None, "id"),
            stanza.IQ.id_.tag)

    def test_tag(self):
        self.assertEqual(
            ("jabber:client", "iq"),
            stanza.IQ.TAG)

    def test_type_attr(self):
        self.assertIsInstance(
            stanza.IQ.type_,
            xso.Attr
        )
        self.assertEqual(
            (None, "type"),
            stanza.IQ.type_.tag
        )
        self.assertIsInstance(
            stanza.IQ.type_.type_,
            xso.EnumCDataType
        )
        self.assertIs(
            stanza.IQ.type_.type_.enum_class,
            structs.IQType,
        )

    def test_error(self):
        self.assertIsInstance(
            stanza.IQ.error,
            xso.Child)

    def test_payload(self):
        self.assertIsInstance(
            stanza.IQ.payload,
            xso.Child)
        self.assertIsNone(stanza.IQ.payload.default)

    def test_reject_init_without_type(self):
        with self.assertRaisesRegex(TypeError, "type_"):
            stanza.IQ()

    def test_init(self):
        payload = TestPayload()

        s = stanza.IQ(
            from_=TEST_FROM,
            type_=structs.IQType.RESULT,
            payload=payload)
        self.assertEqual(
            TEST_FROM,
            s.from_)
        self.assertEqual(
            structs.IQType.RESULT,
            s.type_)
        self.assertIs(
            payload,
            s.payload)

    def test_init_error(self):
        error = object()

        s = stanza.IQ(
            from_=TEST_FROM,
            type_=structs.IQType.ERROR,
            error=error)
        self.assertEqual(
            structs.IQType.ERROR,
            s.type_)
        self.assertIs(
            error,
            s.error)

    def test_make_reply(self):
        s = stanza.IQ(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.IQType.GET)

        r1 = s.make_reply(structs.IQType.ERROR)
        self.assertEqual(
            s.from_,
            r1.to)
        self.assertEqual(
            s.to,
            r1.from_)
        self.assertEqual(
            s.id_,
            r1.id_)
        self.assertEqual(
            structs.IQType.ERROR,
            r1.type_)

    def test_make_reply_enforces_request(self):
        s = stanza.IQ(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.IQType.ERROR)
        with self.assertRaisesRegex(
                ValueError,
                r"make_reply requires request IQ"):
            s.make_reply(unittest.mock.sentinel.type_)
        s.type_ = structs.IQType.RESULT
        with self.assertRaisesRegex(
                ValueError,
                r"make_reply requires request IQ"):
            s.make_reply(unittest.mock.sentinel.type_)

    def test_make_error(self):
        e = stanza.Error(
            condition=errors.ErrorCondition.BAD_REQUEST
        )
        s = stanza.IQ(from_=TEST_FROM,
                      to=TEST_TO,
                      id_="someid",
                      type_=structs.IQType.GET)
        r = s.make_error(e)

        self.assertIsInstance(r, stanza.IQ)

        self.assertEqual(
            r.type_,
            structs.IQType.ERROR)
        self.assertEqual(
            TEST_FROM,
            r.to)
        self.assertEqual(
            TEST_TO,
            r.from_)
        self.assertEqual(
            s.id_,
            r.id_)

    def test_repr(self):
        s = stanza.IQ(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.IQType.ERROR)
        s.error = stanza.Error()
        self.assertEqual(
            "<iq from='foo@example.test' to='bar@example.test'"
            " id='someid' type=<IQType.ERROR: 'error'>"
            " error=<undefined-condition type=<ErrorType.CANCEL: 'cancel'>>>",
            repr(s)
        )

        s = stanza.IQ(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.IQType.RESULT)
        s.payload = TestPayload()
        self.assertEqual(
            "<iq from='foo@example.test' to='bar@example.test'"
            " id='someid' type=<IQType.RESULT: 'result'>"
            " data=foobar>",
            repr(s)
        )

        s = stanza.IQ(
            from_=TEST_FROM,
            to=TEST_TO,
            id_="someid",
            type_=structs.IQType.RESULT)
        self.assertEqual(
            "<iq from='foo@example.test' to='bar@example.test'"
            " id='someid' type=<IQType.RESULT: 'result'>>",
            repr(s)
        )

    def test_repr_works_with_mostly_uninitialised_attributes(self):
        s = stanza.IQ.__new__(stanza.IQ)
        self.assertEqual(
            repr(s),
            "<iq from=None to=None id=<unset> type=<unset> "
            "error=None data=None>"
        )

    def test_repr_works_with_incomplete_attributes(self):
        s = stanza.IQ.__new__(stanza.IQ)
        stanza.IQ.from_.mark_incomplete(s)
        stanza.IQ.to.mark_incomplete(s)
        stanza.IQ.type_.mark_incomplete(s)
        stanza.IQ.id_.mark_incomplete(s)
        self.assertEqual(
            repr(s),
            "<iq from=<incomplete> to=<incomplete> id=<incomplete> type=<incomplete> "
            "error=None data=None>"
        )

    def test__validate_requires_id(self):
        iq = stanza.IQ(structs.IQType.GET)
        with self.assertRaisesRegex(
                ValueError,
                "IQ requires ID"):
            iq._validate()

    def test_as_payload_class(self):
        @stanza.IQ.as_payload_class
        class Foo(xso.XSO):
            TAG = ("uri:foo", "test_as_payload_class")

        self.assertIn(Foo.TAG, stanza.IQ.CHILD_MAP)
        self.assertIs(
            stanza.IQ.CHILD_MAP[Foo.TAG],
            stanza.IQ.payload.xq_descriptor
        )

    def test__validate_rejects_error_without_error(self):
        iq = stanza.IQ(structs.IQType.ERROR)
        iq.autoset_id()

        with self.assertRaisesRegex(
                ValueError,
                r"IQ with type='error' requires error payload"):
            iq._validate()

    def test_validate_wraps_exceptions_from__validate(self):
        class FooException(Exception):
            pass

        iq = stanza.IQ(structs.IQType.GET)

        with self.assertRaisesRegex(
                stanza.StanzaError,
                r"invalid IQ stanza"):
            iq.validate()


class Testmake_application_error(unittest.TestCase):
    def setUp(self):
        self._stack_ctx = contextlib.ExitStack()
        self._stack = self._stack_ctx.__enter__()
        self._as_application_condition = self._stack.enter_context(
            unittest.mock.patch.object(stanza.Error,
                                       "as_application_condition")
        )

    def test_creates_xso_class(self):
        Cls = stanza.make_application_error(
            "TestError",
            ("uri:foo", "bar"),
        )
        self.assertTrue(issubclass(Cls, xso.XSO))
        self.assertIsInstance(Cls, xso.model.XMLStreamClass)
        self.assertEqual(Cls.TAG, ("uri:foo", "bar"))
        self.assertEqual(Cls.__name__, "TestError")

    def test_registers_class(self):
        Cls = stanza.make_application_error(
            "TestError",
            ("uri:foo", "bar"),
        )
        self.assertSequenceEqual(
            self._as_application_condition.mock_calls,
            [
                unittest.mock.call(Cls)
            ]
        )

    def tearDown(self):
        self._stack_ctx.__exit__(None, None, None)
        del self._stack
        del self._stack_ctx



########################################################################
# File name: test_structs.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import collections.abc
import enum
import unittest
import warnings

import aioxmpp
import aioxmpp.structs as structs
import aioxmpp.stanza as stanza


class DisableCompat:
    def __enter__(self):
        if aioxmpp.version_info < (1, 0, 0):
            structs._USE_COMPAT_ENUM = False

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if aioxmpp.version_info < (1, 0, 0):
            structs._USE_COMPAT_ENUM = True


class TestCompatibilityMixin(unittest.TestCase):
    class SomeEnum(structs.CompatibilityMixin, enum.Enum):
        X = "foo"
        Y = "bar"
        Z = None

    def test_compares_normally_without_warnings(self):
        E = self.SomeEnum

        with warnings.catch_warnings(record=True) as w:
            self.assertEqual(E.X, E.X)
            self.assertNotEqual(E.X, E.Y)
            self.assertNotEqual(E.X, E.Z)

            self.assertNotEqual(E.Y, E.X)
            self.assertEqual(E.Y, E.Y)
            self.assertNotEqual(E.Y, E.Z)

            self.assertNotEqual(E.Z, E.X)
            self.assertNotEqual(E.Z, E.Y)
            self.assertEqual(E.Z, E.Z)

        self.assertFalse(w)

    def test_hashes_to_values(self):
        for member in self.SomeEnum:
            self.assertEqual(
                hash(member),
                hash(member.value)
            )

    def _test_eq_with_warning(self, v1, v2):
        with self.assertWarnsRegex(
                DeprecationWarning,
                r"as of aioxmpp 1.0, SomeEnum members will not compare equal to"
                r" their values") as ctx:
            self.assertTrue(v1 == v2)

        self.assertIn(
            "test_structs.py",
            ctx.filename,
        )

        with self.assertWarnsRegex(
                DeprecationWarning,
                r"as of aioxmpp 1.0, SomeEnum members will not compare equal to"
                r" their values") as ctx:
            self.assertFalse(v1 != v2)

        self.assertIn(
            "test_structs.py",
            ctx.filename,
        )

    @unittest.skipIf(aioxmpp.version_info >= (1, 0, 0),
                     "does not apply to this version of aioxmpp")
    def test_compares_equal_to_values_with_DeprecationWarning(self):
        for member in self.SomeEnum:
            self._test_eq_with_warning(member, member.value)
            self._test_eq_with_warning(member.value, member)

    @unittest.skipUnless(aioxmpp.version_info >= (1, 0, 0),
                         "does not apply to this version of aioxmpp")
    def test_compares_not_equal_to_values_by_default(self):
        for member in self.SomeEnum:
            self.assertTrue(member != member.value)
            self.assertTrue(member.value != member)
            self.assertFalse(member == member.value)
            self.assertFalse(member.value == member)

    def test_compares_not_equal_to_values_with_compat_disabled(self):
        with DisableCompat():
            for member in self.SomeEnum:
                self.assertTrue(member != member.value)
                self.assertTrue(member.value != member)
                self.assertFalse(member == member.value)
                self.assertFalse(member.value == member)


class TestErrorType(unittest.TestCase):
    @unittest.skipIf(aioxmpp.version_info >= (1, 0, 0),
                     "does not apply to this version of aioxmpp")
    def test_uses_compat_mixin(self):
        self.assertTrue(
            issubclass(
                structs.ErrorType,
                structs.CompatibilityMixin,
            )
        )


class TestMessageType(unittest.TestCase):
    @unittest.skipIf(aioxmpp.version_info >= (1, 0, 0),
                     "does not apply to this version of aioxmpp")
    def test_uses_compat_mixin(self):
        self.assertTrue(
            issubclass(
                structs.MessageType,
                structs.CompatibilityMixin,
            )
        )

    def test_values(self):
        self.assertSetEqual(
            {v.value for v in structs.MessageType},
            {
                "chat",
                "normal",
                "headline",
                "groupchat",
                "error",
            }
        )

    def test_is_response(self):
        for member in structs.MessageType:
            self.assertEqual(
                member == structs.MessageType.ERROR,
                member.is_response,
            )

    def test_is_error(self):
        for member in structs.MessageType:
            self.assertEqual(
                member == structs.MessageType.ERROR,
                member.is_error,
            )

    def test_is_request(self):
        for member in structs.MessageType:
            self.assertFalse(member.is_request)


class TestPresenceType(unittest.TestCase):
    @unittest.skipIf(aioxmpp.version_info >= (1, 0, 0),
                     "does not apply to this version of aioxmpp")
    def test_uses_compat_mixin(self):
        self.assertTrue(
            issubclass(
                structs.PresenceType,
                structs.CompatibilityMixin,
            )
        )

    def test_values(self):
        self.assertSetEqual(
            {v.value for v in structs.PresenceType},
            {
                "error",
                "probe",
                "subscribe",
                "subscribed",
                "unsubscribe",
                "unsubscribed",
                "unavailable",
                None,
            }
        )

    def test_is_response(self):
        for member in structs.PresenceType:
            self.assertEqual(
                member == structs.PresenceType.ERROR,
                member.is_response,
            )

    def test_is_error(self):
        for member in structs.PresenceType:
            self.assertEqual(
                member == structs.PresenceType.ERROR,
                member.is_error,
            )

    def test_is_request(self):
        for member in structs.PresenceType:
            self.assertFalse(member.is_request)

    def test_is_presence_state(self):
        positive = [
            structs.PresenceType.AVAILABLE,
            structs.PresenceType.UNAVAILABLE,
        ]

        for member in structs.PresenceType:
            self.assertEqual(
                member in positive,
                member.is_presence_state,
            )


class TestIQType(unittest.TestCase):
    @unittest.skipIf(aioxmpp.version_info >= (1, 0, 0),
                     "does not apply to this version of aioxmpp")
    def test_uses_compat_mixin(self):
        self.assertTrue(
            issubclass(
                structs.IQType,
                structs.CompatibilityMixin,
            )
        )

    def test_values(self):
        self.assertSetEqual(
            {member.value for member in structs.IQType},
            {
                "get",
                "set",
                "result",
                "error",
            }
        )

    def test_is_error(self):
        for member in structs.IQType:
            self.assertEqual(
                member == structs.IQType.ERROR,
                member.is_error
            )

    def test_is_request(self):
        positive = [
            structs.IQType.GET,
            structs.IQType.SET,
        ]

        for member in structs.IQType:
            self.assertEqual(
                member in positive,
                member.is_request,
            )

    def test_is_response(self):
        positive = [
            structs.IQType.ERROR,
            structs.IQType.RESULT,
        ]

        for member in structs.IQType:
            self.assertEqual(
                member in positive,
                member.is_response,
            )


class TestJID(unittest.TestCase):
    def test_init_full(self):
        j = structs.JID("foo", "example.com", "bar")
        self.assertEqual(
            "foo",
            j.localpart)
        self.assertEqual(
            "example.com",
            j.domain)
        self.assertEqual(
            "bar",
            j.resource)

    def test_init_enforces_stringprep(self):
        with self.assertRaises(ValueError):
            structs.JID("\u0007", "example.com", "bar")
        with self.assertRaises(ValueError):
            structs.JID("foo", "\u070f", "bar")
        with self.assertRaises(ValueError):
            structs.JID("foo", "example.com", "\u0007")

        self.assertEqual(
            "ssa",
            structs.JID("ßA", "example.test", None).localpart)

        self.assertEqual(
            "ix.test",
            structs.JID(None, "IX.test", None).domain)

        self.assertEqual(
            "IX",
            structs.JID(None, "example.test", "\u2168").resource)

    def test_init_with_default_strict_errors_on_unassigned(self):
        with self.assertRaises(ValueError):
            structs.JID("\U0001f601", "example.com", "bar")
        with self.assertRaises(ValueError):
            structs.JID("foo", "\U0001f601example.com", "bar")
        with self.assertRaises(ValueError):
            structs.JID("foo", "example.com", "\U0001f601")

    def test_init_without_strict_does_not_error_on_unassigned(self):
        structs.JID("\U0001f601", "example.com", "bar", strict=False)
        structs.JID("foo", "\U0001f601example.com", "bar", strict=False)
        structs.JID("foo", "example.com", "\U0001f601", strict=False)

    def test_replace(self):
        j = structs.JID("foo", "example.com", "bar")
        j2 = j.replace(localpart="fnord",
                       domain="example.invalid",
                       resource="baz")
        self.assertEqual(
            "fnord",
            j2.localpart)
        self.assertEqual(
            "example.invalid",
            j2.domain)
        self.assertEqual(
            "baz",
            j2.resource)

    def test_replace_enforces_stringprep(self):
        j = structs.JID("foo", "example.com", "bar")
        with self.assertRaises(ValueError):
            j.replace(localpart="\u0007")
        with self.assertRaises(ValueError):
            j.replace(domain="\u070f")
        with self.assertRaises(ValueError):
            j.replace(resource="\u0007")

        self.assertEqual(
            "ssa",
            j.replace(localpart="ßA").localpart)
        self.assertEqual(
            "ix.test",
            j.replace(domain="IX.test").domain)
        self.assertEqual(
            "IX",
            j.replace(resource="\u2168").resource)

    def test_hashable(self):
        j1 = structs.JID("foo", "bar", "baz")
        j2 = structs.JID("foo", "bar", "baz")
        self.assertEqual(
            hash(j1),
            hash(j2))

    def test_eq(self):
        j1 = structs.JID("foo", "bar", "baz")
        j2 = structs.JID("foo", "bar", "baz")
        self.assertEqual(j1, j2)

    def test_ne(self):
        j1 = structs.JID("foo", "bar", "baz")
        j2 = structs.JID("fooo", "bar", "baz")
        self.assertNotEqual(j1, j2)

    def test_str_full_jid(self):
        j = structs.JID("foo", "example.test", "bar")
        self.assertEqual(
            "foo@example.test/bar",
            str(j))

    def test_str_bare_jid(self):
        j = structs.JID("foo", "example.test", None)
        self.assertEqual(
            "foo@example.test",
            str(j))

    def test_str_domain_jid(self):
        j = structs.JID(None, "example.test", None)
        self.assertEqual(
            "example.test",
            str(j))

    def test_init_bare_jid(self):
        j = structs.JID("foo", "example.test", None)
        self.assertIsNone(j.resource)
        self.assertEqual(
            "foo",
            j.localpart)
        self.assertEqual(
            "example.test",
            j.domain)

    def test_init_domain_jid(self):
        j = structs.JID(None, "example.test", None)
        self.assertIsNone(j.resource)
        self.assertIsNone(j.localpart)
        self.assertEqual(
            "example.test",
            j.domain)

    def test_replace_domain_jid(self):
        j = structs.JID("foo", "example.test", "bar")
        self.assertEqual(
            structs.JID(None, "example.test", None),
            j.replace(localpart=None, resource=None)
        )

    def test_replace_require_domainpart(self):
        j = structs.JID("foo", "example.test", "bar")
        with self.assertRaises(ValueError):
            j.replace(domain=None)

    def test_require_domainpart(self):
        with self.assertRaises(ValueError):
            structs.JID(None, None, None)

    def test_replace_rejects_surplus_argument(self):
        j = structs.JID("foo", "example.test", "bar")
        with self.assertRaises(TypeError):
            j.replace(foobar="baz")

    def test_replace_ignores_problems_on_existing_parts(self):
        j = structs.JID(
            "\U0001f601foo", "\U0001f601example.test", "\U0001f601bar",
            strict=False,
        )

        j2 = j.replace()
        self.assertEqual(j, j2)

    def test_replace_checks_replaced_strings(self):
        j = structs.JID(
            "\U0001f601foo", "\U0001f601example.test", "\U0001f601bar",
            strict=False,
        )

        with self.assertRaises(ValueError):
            j.replace(
                domain=j.domain
            )

    def test_replace_nonstrict_allows_unassigned_codepoints(self):
        j = structs.JID(
            "\U0001f601foo", "\U0001f601example.test", "\U0001f601bar",
            strict=False,
        )

        j2 = j.replace(
            domain=j.domain,
            strict=False,
        )

        self.assertEqual(j, j2)

    def test_immutable(self):
        j = structs.JID(None, "example.test", None)
        with self.assertRaises(AttributeError):
            j.foo = "bar"

    def test_bare(self):
        j = structs.JID("foo", "example.test", "bar")
        self.assertEqual(
            structs.JID("foo", "example.test", None),
            j.bare())

    def test_is_bare(self):
        self.assertFalse(structs.JID("foo", "example.test", "bar").is_bare)
        self.assertTrue(structs.JID("foo", "example.test", None).is_bare)
        self.assertTrue(structs.JID(None, "example.test", None).is_bare)

    def test_is_domain(self):
        self.assertFalse(structs.JID("foo", "example.test", "bar").is_domain)
        self.assertFalse(structs.JID("foo", "example.test", None).is_domain)
        self.assertTrue(structs.JID(None, "example.test", None).is_domain)

    def test_fromstr_full(self):
        self.assertEqual(
            structs.JID("foo", "example.test", "bar"),
            structs.JID.fromstr("foo@example.test/bar")
        )
        self.assertEqual(
            structs.JID("ßA", "IX.test", "\u2168"),
            structs.JID.fromstr("ssa@ix.test/IX")
        )
        self.assertEqual(
            structs.JID("ßA", "IX.test", "bar@baz/fnord"),
            structs.JID.fromstr("ssa@ix.test/bar@baz/fnord")
        )

    def test_fromstr_bare(self):
        self.assertEqual(
            structs.JID("foo", "example.test", None),
            structs.JID.fromstr("foo@example.test")
        )
        self.assertEqual(
            structs.JID("ßA", "IX.test", None),
            structs.JID.fromstr("ssa@ix.test")
        )

    def test_fromstr_domain(self):
        self.assertEqual(
            structs.JID(None, "example.test", None),
            structs.JID.fromstr("example.test")
        )
        self.assertEqual(
            structs.JID(None, "IX.test", None),
            structs.JID.fromstr("ix.test")
        )

    def test_fromstr_domain_with_funny_resource(self):
        self.assertEqual(
            structs.JID(None, "example.test", "foo@bar"),
            structs.JID.fromstr("example.test/foo@bar")
        )

    def test_fromstr_domain_nonstrict(self):
        self.assertEqual(
            structs.JID("\U0001f601", "\U0001f601example.test", "\U0001f601",
                        strict=False),
            structs.JID.fromstr("\U0001f601@\U0001f601example.test/\U0001f601",
                                strict=False)
        )

    def test_reject_empty_localpart(self):
        with self.assertRaises(ValueError):
            structs.JID("", "bar.baz", None)
        with self.assertRaises(ValueError):
            structs.JID.fromstr("@bar.baz")

    def test_reject_empty_domainpart(self):
        with self.assertRaises(ValueError):
            structs.JID("foo", "", None)
        with self.assertRaises(ValueError):
            structs.JID.fromstr("foo@")

    def test_reject_empty_resource(self):
        with self.assertRaises(ValueError):
            structs.JID("foo", "bar.baz", "")
        with self.assertRaises(ValueError):
            structs.JID.fromstr("foo@bar.baz/")

    def test_reject_long_localpart(self):
        with self.assertRaisesRegex(ValueError, "too long"):
            structs.JID("x"*1024, "foo", None)
        with self.assertRaisesRegex(ValueError, "too long"):
            structs.JID("ü"*512, "foo", None)
        with self.assertRaisesRegex(ValueError, "too long"):
            structs.JID.fromstr("ü"*512 + "@foo")

    def test_reject_long_domainpart(self):
        with self.assertRaisesRegex(ValueError, "too long"):
            structs.JID(None, "x"*1024, None)
        with self.assertRaisesRegex(ValueError, "too long"):
            structs.JID(None, "ü"*512, None)
        with self.assertRaisesRegex(ValueError, "too long"):
            structs.JID.fromstr("ü"*512)

    def test_reject_long_resource(self):
        with self.assertRaisesRegex(ValueError, "too long"):
            structs.JID(None, "foo", "x"*1024)
        with self.assertRaisesRegex(ValueError, "too long"):
            structs.JID(None, "foo", "ü"*512)
        with self.assertRaisesRegex(ValueError, "too long"):
            structs.JID.fromstr("foo/" + "ü"*512)


class TestPresenceShow(unittest.TestCase):
    def test_aliases(self):
        self.assertIs(
            structs.PresenceShow.XA,
            structs.PresenceShow.EXTENDED_AWAY
        )

        self.assertIs(
            structs.PresenceShow.PLAIN,
            structs.PresenceShow.NONE
        )

        self.assertIs(
            structs.PresenceShow.CHAT,
            structs.PresenceShow.FREE_FOR_CHAT
        )

        self.assertIs(
            structs.PresenceShow.DND,
            structs.PresenceShow.DO_NOT_DISTURB
        )

    def test_ordering_simple(self):
        values = [
            structs.PresenceShow.AWAY,
            structs.PresenceShow.CHAT,
            structs.PresenceShow.PLAIN,
            structs.PresenceShow.DND,
            structs.PresenceShow.XA,
        ]
        values.sort()

        self.assertSequenceEqual(
            [
                structs.PresenceShow.XA,
                structs.PresenceShow.AWAY,
                structs.PresenceShow.PLAIN,
                structs.PresenceShow.CHAT,
                structs.PresenceShow.DND,
            ],
            values,
        )

    def test_proper_error_message_on_invalid_ordering_operand(self):
        with self.assertRaises(TypeError):
            structs.PresenceShow.AWAY < 1

    def test_value(self):
        values = [
            "xa",
            "away",
            None,
            "chat",
            "dnd"
        ]

        for v in values:
            m = structs.PresenceShow(v)
            self.assertEqual(m.value, v)

    def test_ordering(self):
        values = [
            structs.PresenceShow("xa"),
            structs.PresenceShow("away"),
            structs.PresenceShow(None),
            structs.PresenceShow("chat"),
            structs.PresenceShow("dnd"),
        ]

        for i in range(1, len(values)-1):
            for v1, v2 in zip(values[:-i], values[i:]):
                self.assertLess(v1, v2)
                self.assertLessEqual(v1, v2)
                self.assertNotEqual(v1, v2)
                self.assertGreater(v2, v1)
                self.assertGreaterEqual(v2, v1)

    @unittest.skipIf(aioxmpp.version_info >= (1, 0, 0),
                     "does not apply to this version of aioxmpp")
    def test_uses_compat_mixin(self):
        self.assertTrue(
            issubclass(
                structs.PresenceShow,
                structs.CompatibilityMixin,
            )
        )


class TestPresenceState(unittest.TestCase):
    def test_immutable(self):
        ps = structs.PresenceState()
        with self.assertRaises(AttributeError):
            ps.foo = "bar"
        with self.assertRaises(AttributeError):
            ps.available = True
        with self.assertRaises(AttributeError):
            ps.show = "baz"

    def test_init_defaults(self):
        ps = structs.PresenceState()
        self.assertFalse(ps.available)
        self.assertEqual(ps.show, structs.PresenceShow.NONE)

    def test_init_compat(self):
        with self.assertWarnsRegex(
                DeprecationWarning,
                "as of aioxmpp 1.0, the show argument must use "
                "PresenceShow instead of str") as ctx:
            ps = structs.PresenceState(True, "dnd")
        self.assertIn(
            "test_structs.py",
            ctx.filename,
        )
        self.assertTrue(ps.available)
        self.assertEqual(ps.show, structs.PresenceShow.DND)

    def test_init_available(self):
        ps = structs.PresenceState(available=True)
        self.assertTrue(ps.available)

    def test_init_normalizes_available(self):
        ps = structs.PresenceState(available="foo")
        self.assertIs(True, ps.available)

    def test_init_available_with_show(self):
        ps = structs.PresenceState(available=True,
                                   show=structs.PresenceShow.DND)
        self.assertTrue(ps.available)
        self.assertIs(structs.PresenceShow.DND, ps.show)

    def test_init_available_validate_show(self):
        with self.assertRaises(ValueError):
            ps = structs.PresenceState(available=True, show="foobar")
        for value in ["dnd", "xa", "away", None, "chat"]:
            value = structs.PresenceShow(value)
            ps = structs.PresenceState(
                available=True,
                show=value)
            self.assertEqual(value, ps.show)

    def test_init_unavailable_forbids_show(self):
        with self.assertRaises(ValueError):
            structs.PresenceState(available=False,
                                  show=structs.PresenceShow.DND)

    def test_ordering(self):
        values = [
            structs.PresenceState(),
            structs.PresenceState(available=True,
                                  show=structs.PresenceShow.XA),
            structs.PresenceState(available=True,
                                  show=structs.PresenceShow.AWAY),
            structs.PresenceState(available=True),
            structs.PresenceState(available=True,
                                  show=structs.PresenceShow.CHAT),
            structs.PresenceState(available=True,
                                  show=structs.PresenceShow.DND),
        ]

        for i in range(1, len(values)-1):
            for v1, v2 in zip(values[:-i], values[i:]):
                self.assertLess(v1, v2)
                self.assertLessEqual(v1, v2)
                self.assertNotEqual(v1, v2)
                self.assertGreater(v2, v1)
                self.assertGreaterEqual(v2, v1)

    def test_proper_exception_on_invalid_ordering_operand(self):
        with self.assertRaises(TypeError):
            structs.PresenceState() < 1

        with self.assertRaises(TypeError):
            structs.PresenceState() > 1

        with self.assertRaises(TypeError):
            structs.PresenceState() >= 1

        with self.assertRaises(TypeError):
            structs.PresenceState() <= 1

        self.assertFalse(structs.PresenceState() == 0)
        self.assertTrue(structs.PresenceState() != 0)

    def test_equality(self):
        self.assertEqual(
            structs.PresenceState(),
            structs.PresenceState()
        )
        self.assertEqual(
            structs.PresenceState(available=True),
            structs.PresenceState(available=True)
        )
        self.assertEqual(
            structs.PresenceState(available=True,
                                  show=structs.PresenceShow.DND),
            structs.PresenceState(available=True,
                                  show=structs.PresenceShow.DND),
        )
        self.assertFalse(
            structs.PresenceState(available=True,
                                  show=structs.PresenceShow.DND) !=
            structs.PresenceState(available=True,
                                  show=structs.PresenceShow.DND)
        )

    def test_equality_deals_with_different_types(self):
        self.assertNotEqual(structs.PresenceState(), None)
        self.assertNotEqual(structs.PresenceState(), "foo")
        self.assertNotEqual(structs.PresenceState(), 123)

    def test_repr(self):
        self.assertEqual(
            "<PresenceState>",
            repr(structs.PresenceState())
        )
        self.assertEqual(
            "<PresenceState available>",
            repr(structs.PresenceState(available=True))
        )
        self.assertEqual(
            "<PresenceState available show=<PresenceShow.DND: 'dnd'>>",
            repr(structs.PresenceState(available=True,
                                       show=structs.PresenceShow.DND))
        )

    def test_apply_to_stanza(self):
        stanza_obj = stanza.Presence(type_=structs.PresenceType.PROBE)
        self.assertEqual(stanza_obj.show, structs.PresenceShow.NONE)

        ps = structs.PresenceState(available=True,
                                   show=structs.PresenceShow.DND)
        ps.apply_to_stanza(stanza_obj)
        self.assertEqual(
            structs.PresenceType.AVAILABLE,
            stanza_obj.type_
        )
        self.assertEqual(
            structs.PresenceShow.DND,
            stanza_obj.show
        )

        ps = structs.PresenceState()
        ps.apply_to_stanza(stanza_obj)
        self.assertEqual(
            structs.PresenceType.UNAVAILABLE,
            stanza_obj.type_
        )
        self.assertEqual(
            stanza_obj.show,
            structs.PresenceShow.NONE,
        )

    def test_from_stanza(self):
        stanza_obj = stanza.Presence(
            type_=structs.PresenceType.AVAILABLE
        )
        stanza_obj.show = structs.PresenceShow.XA
        self.assertEqual(
            structs.PresenceState(available=True,
                                  show=structs.PresenceShow.XA),
            structs.PresenceState.from_stanza(stanza_obj)
        )

        stanza_obj = stanza.Presence(
            type_=structs.PresenceType.UNAVAILABLE,
        )
        self.assertEqual(
            structs.PresenceState(available=False),
            structs.PresenceState.from_stanza(stanza_obj)
        )

    def test_from_stanza_reject_incorrect_types(self):
        stanza_obj = stanza.Presence(
            type_=structs.PresenceType.PROBE
        )
        with self.assertRaises(ValueError):
            structs.PresenceState.from_stanza(stanza_obj)

    def test_from_stanza_nonstrict_by_default(self):
        stanza_obj = stanza.Presence(
            type_=structs.PresenceType.UNAVAILABLE
        )
        stanza_obj.show = structs.PresenceShow.AWAY
        self.assertEqual(
            structs.PresenceState(available=False),
            structs.PresenceState.from_stanza(stanza_obj)
        )

    def test_from_stanza_strict_by_default(self):
        stanza_obj = stanza.Presence(
            type_=structs.PresenceType.UNAVAILABLE,
        )
        stanza_obj.show = structs.PresenceShow.DND
        with self.assertRaises(ValueError):
            structs.PresenceState.from_stanza(stanza_obj, strict=True)


class TestLanguageTag(unittest.TestCase):
    def test_init_requires_kwargs(self):
        with self.assertRaisesRegex(TypeError,
                                     "takes 1 positional argument"):
            structs.LanguageTag("foo")

    def test_init_requires_language(self):
        with self.assertRaisesRegex(ValueError, "tag cannot be empty"):
            structs.LanguageTag()

    def test_fromstr_match_str(self):
        tag = structs.LanguageTag.fromstr("de-Latn-DE-1999")
        self.assertEqual(
            "de-latn-de-1999",
            tag.match_str
        )

    def test_fromstr_print_str(self):
        tag = structs.LanguageTag.fromstr("de-Latn-DE-1999")
        self.assertEqual(
            "de-Latn-DE-1999",
            tag.print_str
        )

    def test___str__(self):
        tag = structs.LanguageTag.fromstr("zh-Hans")
        self.assertEqual(
            "zh-Hans",
            str(tag)
        )
        tag = structs.LanguageTag.fromstr("de-Latn-DE-1999")
        self.assertEqual(
            "de-Latn-DE-1999",
            str(tag)
        )

    def test_compare_case_insensitively(self):
        tag1 = structs.LanguageTag.fromstr("de-DE")
        tag2 = structs.LanguageTag.fromstr("de-de")
        tag3 = structs.LanguageTag.fromstr("fr")

        self.assertTrue(tag1 == tag2)
        self.assertFalse(tag1 != tag2)
        self.assertTrue(tag2 == tag1)
        self.assertFalse(tag2 != tag1)

        self.assertTrue(tag1 != tag3)
        self.assertFalse(tag1 == tag3)
        self.assertTrue(tag2 != tag3)
        self.assertFalse(tag2 == tag3)

        self.assertTrue(tag3 != tag1)
        self.assertFalse(tag3 == tag1)
        self.assertTrue(tag3 != tag1)
        self.assertFalse(tag3 == tag1)

    def test_order_case_insensitively(self):
        tag1 = structs.LanguageTag.fromstr("de-DE")
        tag2 = structs.LanguageTag.fromstr("de-de")
        tag3 = structs.LanguageTag.fromstr("en-us")
        tag4 = structs.LanguageTag.fromstr("fr")

        self.assertLess(tag1, tag3)
        self.assertLess(tag1, tag4)
        self.assertLess(tag2, tag3)
        self.assertLess(tag2, tag4)
        self.assertLess(tag3, tag4)

        self.assertGreater(tag4, tag3)
        self.assertGreater(tag4, tag2)
        self.assertGreater(tag4, tag1)
        self.assertGreater(tag3, tag2)
        self.assertGreater(tag3, tag1)

        self.assertFalse(tag1 > tag2)
        self.assertFalse(tag2 > tag1)

        self.assertFalse(tag1 < tag2)
        self.assertFalse(tag2 < tag1)

    def test_hash_case_insensitively(self):
        tag1 = structs.LanguageTag.fromstr("de-DE")
        tag2 = structs.LanguageTag.fromstr("de-de")

        self.assertEqual(hash(tag1), hash(tag2))

    def test_not_equal_to_None(self):
        tag1 = structs.LanguageTag.fromstr("de-DE")
        self.assertNotEqual(tag1, None)

    def test_dont_compare_with_None(self):
        tag1 = structs.LanguageTag.fromstr("de-DE")
        with self.assertRaises(TypeError):
            tag1 > None
        with self.assertRaises(TypeError):
            tag1 < None
        with self.assertRaises(TypeError):
            tag1 >= None
        with self.assertRaises(TypeError):
            tag1 <= None

    def test__repr__(self):
        tag1 = structs.LanguageTag.fromstr("de-DE")
        tag2 = structs.LanguageTag.fromstr("fr")

        self.assertEqual(
            "<aioxmpp.structs.LanguageTag.fromstr('de-DE')>",
            repr(tag1)
        )

        self.assertEqual(
            "<aioxmpp.structs.LanguageTag.fromstr('fr')>",
            repr(tag2)
        )

    def test_immutable(self):
        tag = structs.LanguageTag.fromstr("foo")
        with self.assertRaises(AttributeError):
            tag.foo = "bar"


class TestLanguageRange(unittest.TestCase):
    def test_init_requires_kwargs(self):
        with self.assertRaisesRegex(TypeError,
                                     "takes 1 positional argument"):
            structs.LanguageRange("foo")

    def test_init_requires_language(self):
        with self.assertRaisesRegex(ValueError, "range cannot be empty"):
            structs.LanguageRange()

    def test_fromstr_match_str(self):
        tag = structs.LanguageRange.fromstr("de-DE")
        self.assertEqual(
            "de-de",
            tag.match_str
        )

    def test_fromstr_print_str(self):
        tag = structs.LanguageRange.fromstr("de-Latn-DE-1999")
        self.assertEqual(
            "de-Latn-DE-1999",
            tag.print_str
        )

    def test___str__(self):
        tag = structs.LanguageRange.fromstr("zh-Hans")
        self.assertEqual(
            "zh-Hans",
            str(tag)
        )
        tag = structs.LanguageRange.fromstr("de-Latn-DE-1999")
        self.assertEqual(
            "de-Latn-DE-1999",
            str(tag)
        )

    def test_compare_case_insensitively(self):
        tag1 = structs.LanguageRange.fromstr("de-DE")
        tag2 = structs.LanguageRange.fromstr("de-de")
        tag3 = structs.LanguageRange.fromstr("fr")

        self.assertTrue(tag1 == tag2)
        self.assertFalse(tag1 != tag2)
        self.assertTrue(tag2 == tag1)
        self.assertFalse(tag2 != tag1)

        self.assertTrue(tag1 != tag3)
        self.assertFalse(tag1 == tag3)
        self.assertTrue(tag2 != tag3)
        self.assertFalse(tag2 == tag3)

        self.assertTrue(tag3 != tag1)
        self.assertFalse(tag3 == tag1)
        self.assertTrue(tag3 != tag1)
        self.assertFalse(tag3 == tag1)

    def test_hash_case_insensitively(self):
        tag1 = structs.LanguageRange.fromstr("de-DE")
        tag2 = structs.LanguageRange.fromstr("de-de")

        self.assertEqual(hash(tag1), hash(tag2))

    def test_not_equal_to_None(self):
        r1 = structs.LanguageRange.fromstr("de-DE")
        self.assertNotEqual(r1, None)

    def test_wildcard(self):
        r1 = structs.LanguageRange.fromstr("*")
        r2 = structs.LanguageRange.fromstr("*")
        self.assertIs(r1, r2)

    def test_strip_rightmost(self):
        r = structs.LanguageRange.fromstr("de-Latn-DE-x-foo")
        self.assertEqual(
            structs.LanguageRange.fromstr("de-Latn-DE"),
            r.strip_rightmost()
        )
        self.assertEqual(
            structs.LanguageRange.fromstr("de-Latn"),
            r.strip_rightmost().strip_rightmost()
        )
        self.assertEqual(
            structs.LanguageRange.fromstr("de"),
            r.strip_rightmost().strip_rightmost().strip_rightmost()
        )
        with self.assertRaises(ValueError):
            r.strip_rightmost().strip_rightmost()\
                .strip_rightmost().strip_rightmost()

    def test_immutable(self):
        r = structs.LanguageRange.fromstr("foo")
        with self.assertRaises(AttributeError):
            r.foo = "bar"


class Testbasic_filter_languages(unittest.TestCase):
    def setUp(self):
        self.languages = [
            structs.LanguageTag.fromstr("de-Latn-DE-1999"),
            structs.LanguageTag.fromstr("de-DE"),
            structs.LanguageTag.fromstr("de-Latn"),
            structs.LanguageTag.fromstr("fr-CH"),
            structs.LanguageTag.fromstr("it"),
        ]


    def test_filter(self):
        self.assertSequenceEqual(
            [
                self.languages[0],
                self.languages[1],
                self.languages[2],
            ],
            list(structs.basic_filter_languages(
                self.languages,
                list(map(structs.LanguageRange.fromstr, [
                    "de",
                ]))
            ))
        )

        self.assertSequenceEqual(
            [
                self.languages[1],
            ],
            list(structs.basic_filter_languages(
                self.languages,
                list(map(structs.LanguageRange.fromstr, [
                    "de-DE",
                ]))
            ))
        )

        self.assertSequenceEqual(
            [
                self.languages[0],
                self.languages[2],
            ],
            list(structs.basic_filter_languages(
                self.languages,
                list(map(structs.LanguageRange.fromstr, [
                    "de-Latn",
                ]))
            ))
        )

    def test_filter_no_dupes_and_ordered(self):
        self.assertSequenceEqual(
            [
                self.languages[0],
                self.languages[2],
                self.languages[1],
            ],
            list(structs.basic_filter_languages(
                self.languages,
                list(map(structs.LanguageRange.fromstr, [
                    "de-Latn",
                    "de",
                ]))
            ))
        )

    def test_filter_wildcard(self):
        self.assertSequenceEqual(
            self.languages,
            list(structs.basic_filter_languages(
                self.languages,
                list(map(structs.LanguageRange.fromstr, [
                    "fr",
                    "*",
                ]))
            ))
        )


class Testlookup_language(unittest.TestCase):
    def setUp(self):
        self.languages = [
            structs.LanguageTag.fromstr("de-Latn-DE-1999"),
            structs.LanguageTag.fromstr("fr-CH"),
            structs.LanguageTag.fromstr("it"),
        ]

    def test_match_direct(self):
        self.assertEqual(
            structs.LanguageTag.fromstr("fr-CH"),
            structs.lookup_language(
                self.languages,
                list(map(structs.LanguageRange.fromstr, [
                    "en",
                    "fr-ch",
                    "de-de"
                ]))
            )
        )

        self.assertEqual(
            structs.LanguageTag.fromstr("it"),
            structs.lookup_language(
                self.languages,
                list(map(structs.LanguageRange.fromstr, [
                    "it",
                ]))
            )
        )

    def test_decay(self):
        self.assertEqual(
            structs.LanguageTag.fromstr("de-Latn-DE-1999"),
            structs.lookup_language(
                self.languages,
                list(map(structs.LanguageRange.fromstr, [
                    "de-de",
                    "en-GB",
                    "en"
                ]))
            )
        )

        self.assertEqual(
            structs.LanguageTag.fromstr("fr-CH"),
            structs.lookup_language(
                self.languages,
                list(map(structs.LanguageRange.fromstr, [
                    "fr-FR",
                    "de-DE",
                    "fr",
                ]))
            )
        )

    def test_decay_skips_extension_prefixes_properly(self):
        self.assertEqual(
            structs.LanguageTag.fromstr("de-DE"),
            structs.lookup_language(
                list(map(structs.LanguageTag.fromstr, [
                    "de-DE",
                    "de-x",
                ])),
                list(map(structs.LanguageRange.fromstr, [
                    "de-x-foobar",
                ]))
            )
        )


class TestLanguageMap(unittest.TestCase):
    def test_implements_mapping(self):
        mapping = structs.LanguageMap()
        self.assertIsInstance(
            mapping,
            collections.abc.MutableMapping
        )

    def test_mapping_interface(self):
        key1 = structs.LanguageTag.fromstr("de-DE")
        key2 = structs.LanguageTag.fromstr("en-US")
        key3 = structs.LanguageTag.fromstr("en")

        mapping = structs.LanguageMap()

        self.assertFalse(mapping)
        self.assertEqual(0, len(mapping))

        mapping[key1] = 10

        self.assertIn(key1, mapping)
        self.assertEqual(
            10,
            mapping[key1]
        )

        self.assertSetEqual(
            {key1},
            set(mapping)
        )

        mapping[key2] = 20

        self.assertIn(key2, mapping)
        self.assertEqual(
            20,
            mapping[key2]
        )

        self.assertSetEqual(
            {key1, key2},
            set(mapping)
        )

        key2_prime = structs.LanguageTag.fromstr("en-us")

        self.assertIn(key2_prime, mapping)
        self.assertEqual(
            20,
            mapping[key2_prime]
        )

        self.assertNotIn(key3, mapping)

        del mapping[key1]

        self.assertNotIn(key1, mapping)

        mapping.clear()

        self.assertNotIn(key2, mapping)

    def test_lookup(self):
        key1 = structs.LanguageTag.fromstr("de-DE")
        key2 = structs.LanguageTag.fromstr("en-US")
        key3 = structs.LanguageTag.fromstr("en")

        mapping = structs.LanguageMap()

        mapping[key1] = 10
        mapping[key2] = 20
        mapping[key3] = 30

        self.assertEqual(
            30,
            mapping.lookup([structs.LanguageRange.fromstr("en-GB")])
        )

    def test_values(self):
        key1 = structs.LanguageTag.fromstr("de-DE")
        key2 = structs.LanguageTag.fromstr("en-US")
        key3 = structs.LanguageTag.fromstr("en")

        mapping = structs.LanguageMap()

        mapping[key1] = 10
        mapping[key2] = 20
        mapping[key3] = 30

        self.assertSetEqual(
            {10, 20, 30},
            set(mapping.values())
        )

    def test_keys(self):
        key1 = structs.LanguageTag.fromstr("de-DE")
        key2 = structs.LanguageTag.fromstr("en-US")
        key3 = structs.LanguageTag.fromstr("en")

        mapping = structs.LanguageMap()

        mapping[key1] = 10
        mapping[key2] = 20
        mapping[key3] = 30

        self.assertSetEqual(
            {key1, key2, key3},
            set(mapping.keys())
        )

    def test_items(self):
        key1 = structs.LanguageTag.fromstr("de-DE")
        key2 = structs.LanguageTag.fromstr("en-US")
        key3 = structs.LanguageTag.fromstr("en")

        mapping = structs.LanguageMap()

        mapping[key1] = 10
        mapping[key2] = 20
        mapping[key3] = 30

        self.assertSetEqual(
            {
                (key1, 10),
                (key2, 20),
                (key3, 30),
            },
            set(mapping.items())
        )

    def test_equality(self):
        mapping1 = structs.LanguageMap()
        mapping1[structs.LanguageTag.fromstr("de-de")] = 10
        mapping1[structs.LanguageTag.fromstr("en-US")] = 20

        mapping2 = structs.LanguageMap()
        mapping2[structs.LanguageTag.fromstr("de-DE")] = 10
        mapping2[structs.LanguageTag.fromstr("en-US")] = 20

        mapping3 = structs.LanguageMap()
        mapping3[structs.LanguageTag.fromstr("de-DE")] = 10
        mapping3[structs.LanguageTag.fromstr("en-GB")] = 20

        self.assertEqual(
            mapping1,
            mapping2
        )
        self.assertFalse(mapping1 != mapping2)

        self.assertNotEqual(
            mapping1,
            mapping3
        )
        self.assertFalse(mapping1 == mapping3)

        self.assertNotEqual(
            mapping2,
            mapping3
        )
        self.assertFalse(mapping2 == mapping3)

    def test_setdefault(self):
        l = []
        mapping = structs.LanguageMap()
        result = mapping.setdefault(structs.LanguageTag.fromstr("de-de"), l)

        self.assertIs(result, l)

        result = mapping.setdefault(structs.LanguageTag.fromstr("de-de"), [])

        self.assertIs(result, l)

    def test_lookup_returns_None_key_if_nothing_matches(self):
        mapping = structs.LanguageMap()
        mapping[None] = "foobar"
        mapping[structs.LanguageTag.fromstr("de")] = "Test"
        mapping[structs.LanguageTag.fromstr("en")] = "test"

        self.assertEqual(
            "foobar",
            mapping.lookup([structs.LanguageRange.fromstr("it")])
        )

    def test_any_returns_only_key(self):
        m = structs.LanguageMap()
        m[None] = "fnord"

        self.assertEqual(m.any(), "fnord")

        m = structs.LanguageMap()
        m[structs.LanguageTag.fromstr("de")] = "Test"

        self.assertEqual(m.any(), "Test")

    def test_any_raises_ValueError_on_empty_map(self):
        m = structs.LanguageMap()
        with self.assertRaises(ValueError):
            m.any()

    def test_any_prefers_None(self):
        m = structs.LanguageMap()
        m[structs.LanguageTag.fromstr("de")] = "Test"
        m[None] = "fnord"

        self.assertEqual(m.any(), "fnord")

        m = structs.LanguageMap()
        m[None] = "fnord"
        m[structs.LanguageTag.fromstr("de")] = "Test"

        self.assertEqual(m.any(), "fnord")

    def test_any_returns_same_key_for_same_keyset(self):
        m = structs.LanguageMap()
        m[structs.LanguageTag.fromstr("de")] = "Test"
        m[structs.LanguageTag.fromstr("fr")] = "fnord"

        self.assertEqual(m.any(), "Test")

        m = structs.LanguageMap()
        m[structs.LanguageTag.fromstr("fr")] = "fnord"
        m[structs.LanguageTag.fromstr("de")] = "Test"

        self.assertEqual(m.any(), "Test")



#!/usr/bin/env python3
########################################################################
# File name: block_jid.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import sys

import aioxmpp.disco
import aioxmpp.blocking
import aioxmpp.xso

from framework import Example, exec_example


class BlockJID(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        self.argparse.add_argument(
            "--add",
            dest="jids_to_block",
            default=[],
            action="append",
            type=jid,
            metavar="JID",
            help="JID to block (can be specified multiple times)",
        )

        self.argparse.add_argument(
            "--remove",
            dest="jids_to_unblock",
            default=[],
            action="append",
            type=jid,
            metavar="JID",
            help="JID to unblock (can be specified multiple times)",
        )

        self.argparse.add_argument(
            "-l", "--list",
            action="store_true",
            default=False,
            dest="show_list",
            help="If given, prints the block list at the end of the operation",
        )

    def configure(self):
        super().configure()

        if not (self.args.jids_to_block or
                self.args.show_list or
                self.args.jids_to_unblock):
            print("nothing to do!", file=sys.stderr)
            print("specify --add and/or --remove and/or --list",
                  file=sys.stderr)
            sys.exit(1)

    def make_simple_client(self):
        client = super().make_simple_client()
        self.blocking = client.summon(aioxmpp.BlockingClient)
        return client

    @asyncio.coroutine
    def run_simple_example(self):
        # we are polite and ask the server whether it actually supports the
        # XEP-0191 block list protocol
        disco = self.client.summon(aioxmpp.DiscoClient)
        server_info = yield from disco.query_info(
            self.client.local_jid.replace(
                resource=None,
                localpart=None,
            )
        )

        if "urn:xmpp:blocking" not in server_info.features:
            print("server does not support block lists!", file=sys.stderr)
            sys.exit(2)

        # now that we are sure that the server supports it, we can send
        # requests.

        if self.args.jids_to_block:
            yield from self.blocking.block_jids(self.args.jids_to_block)
        else:
            print("nothing to block")

        if self.args.jids_to_unblock:
            yield from self.blocking.unblock_jids(self.args.jids_to_unblock)
        else:
            print("nothing to unblock")

        if self.args.show_list:
            # print all the items; again, .items is a list of JIDs
            print("current block list:")
            for item in sorted(self.blocking.blocklist):
                print("\t", item, sep="")


if __name__ == "__main__":
    exec_example(BlockJID())



########################################################################
# File name: cache.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.cache` --- Utilities for implementing caches
###########################################################

.. versionadded:: 0.9

    This module was added in version 0.9.

.. autoclass:: LRUDict

"""

import collections.abc


class Node:
    __slots__ = ("prev", "next_", "key", "value")


def _init_linked_list():
    root = Node()
    root.prev = root
    root.next_ = root
    root.key = None
    root.value = None
    return root


def _remove_node(node):
    node.next_.prev = node.prev
    node.prev.next_ = node.next_
    return node


def _insert_node(before, new_node):
    new_node.next_ = before.next_
    new_node.next_.prev = new_node
    new_node.prev = before
    before.next_ = new_node


def _length(node):
    # this is used only for testing
    cur = node.next_
    i = 0
    while cur is not node:
        i += 1
        cur = cur.next_
    return i


def _has_consistent_links(node, node_dict=None):
    # this is used only for testing
    cur = node.next_

    if cur.prev is not node:
        return False

    while cur is not node:
        if node_dict is not None and node_dict[cur.key] is not cur:
            return False
        if cur is not cur.next_.prev:
            return False
        cur = cur.next_
    return True


class LRUDict(collections.abc.MutableMapping):
    """
    Size-restricted dictionary with Least Recently Used expiry policy.

    .. versionadded:: 0.9

    The :class:`LRUDict` supports normal dictionary-style access and implements
    :class:`collections.abc.MutableMapping`.

    When the :attr:`maxsize` is exceeded, as many entries as needed to get
    below the :attr:`maxsize` are removed from the dict. Least recently used
    entries are purged first. Setting an entry does *not* count as use!

    .. autoattribute:: maxsize
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__links = {}
        self.__root = _init_linked_list()

        self.__maxsize = 1

    def _test_consistency(self):
        """
        This method is only used for testing to assert that the operations
        leave the LRUDict in a valid state.
        """
        return (_length(self.__root) == len(self.__links) and
                _has_consistent_links(self.__root, self.__links))

    def _purge(self):
        if self.__maxsize is None:
            return

        while len(self.__links) > self.__maxsize:
            link = _remove_node(self.__root.prev)
            del self.__links[link.key]

    @property
    def maxsize(self):
        """
        Maximum size of the cache. Changing this property purges overhanging
        entries immediately.

        If set to :data:`None`, no limit on the number of entries is imposed.
        Do **not** use a limit of :data:`None` for data where the `key` is
        under control of a remote entity.

        Use cases for :data:`None` are those where you only need the explicit
        expiry feature, but not the LRU feature.
        """
        return self.__maxsize

    @maxsize.setter
    def maxsize(self, value):
        if value is not None and value <= 0:
            raise ValueError("maxsize must be positive integer or None")
        self.__maxsize = value
        self._purge()

    def __len__(self):
        return len(self.__links)

    def __iter__(self):
        return iter(self.__links)

    def __setitem__(self, key, value):
        try:
            self.__links[key].value = value
        except KeyError:
            link = Node()
            link.key = key
            link.value = value
            self.__links[key] = link
            _insert_node(self.__root, link)
            self._purge()

    def __getitem__(self, key):
        link = self.__links[key]
        _remove_node(link)
        _insert_node(self.__root, link)
        return link.value

    def __delitem__(self, key):
        link = self.__links.pop(key)
        _remove_node(link)

    def clear(self):
        self.__links.clear()
        self.__root = _init_linked_list()
